@@ -16,7 +16,6 @@
  *    except according to the terms contained in the LICENSE file.
  */
 
-#define _GNU_SOURCE
 #include <errno.h>
 #include <fcntl.h>
 #include <getopt.h>
@@ -446,7 +445,7 @@ int main(int argc, char **argv)
 	while (optind < argc) {
 		path = clean_path(argv[optind]);
 		if (writable)
-			exit(!is_writable(path));
+			exit(access(path, W_OK) != 0);
 		if (do_check(path, uid, gid, mode, type, trunc, chowner,
 					symlinks, selinux_on))
 			retval = EXIT_FAILURE;
