@@ -4,7 +4,7 @@
 */
 
 /*
- * Copyright (c) 2007-2024 The OpenRC Authors.
+ * Copyright (c) 2007-2025 The OpenRC Authors.
  * See the Authors file at the top-level directory of this distribution and
  * https://github.com/OpenRC/openrc/blob/HEAD/AUTHORS
  *
@@ -31,7 +31,7 @@
 #include "einfo.h"
 #include "helpers.h"
 
-/* Incase we cannot work out how many columns from ioctl, supply a default */
+/* In case we cannot work out how many columns from ioctl, supply a default */
 #define DEFAULT_COLS		 80
 
 #define OK			"ok"
@@ -97,6 +97,7 @@ static bool term_is_cons25 = false;
  * Only terminals without "color" in the name need to be explicitly listed */
 static const char *const color_terms[] = {
 	"Eterm",
+	"alacritty",
 	"ansi",
 	"con132x25",
 	"con132x30",
@@ -459,7 +460,6 @@ _eindent(FILE * EINFO_RESTRICT stream)
 {
 	char *env = getenv("EINFO_INDENT");
 	int amount = 0;
-	char indent[INDENT_MAX];
 
 	if (env) {
 		errno = 0;
@@ -468,14 +468,9 @@ _eindent(FILE * EINFO_RESTRICT stream)
 			amount = 0;
 		else if (amount > INDENT_MAX)
 			amount = INDENT_MAX;
-
-		if (amount > 0)
-			memset(indent, ' ', (size_t)amount);
 	}
 
-	/* Terminate it */
-	memset(indent + amount, 0, 1);
-	return fprintf(stream, "%s", indent);
+	return fprintf(stream, "%*s", amount, "");
 }
 
 static const char *
@@ -793,8 +788,6 @@ eindent(void)
 			amount = 0;
 	}
 	amount += INDENT_WIDTH;
-	if (amount > INDENT_MAX)
-		amount = INDENT_MAX;
 	xasprintf(&num, "%08d", amount);
 	setenv("EINFO_INDENT", num, 1);
 	free(num);
