@@ -101,8 +101,8 @@ rc_find_pids(const char *exec, const char *const *argv, uid_t uid, pid_t pid)
 	bool container_pid = false;
 	bool openvz_host = false;
 	char *line = NULL;
-	char my_ns[30];
-	char proc_ns[30];
+	char my_ns[30] = {0};
+	char proc_ns[30] = {0};
 	size_t len = 0;
 	pid_t p;
 	char *buffer = NULL;
@@ -137,26 +137,18 @@ rc_find_pids(const char *exec, const char *const *argv, uid_t uid, pid_t pid)
 	from our list of pids.
 	*/
 
-	if (exists("/proc/self/status")) {
-		fp = fopen("/proc/self/status", "r");
-		if (fp) {
-			while (xgetline(&line, &len, fp) != -1) {
-				if (strncmp(line, "envID:\t0", 8) == 0) {
-					openvz_host = true;
-					break;
-				}
+	if ((fp = fopen("/proc/self/status", "r"))) {
+		while (xgetline(&line, &len, fp) != -1) {
+			if (strncmp(line, "envID:\t0", 8) == 0) {
+				openvz_host = true;
+				break;
 			}
-			fclose(fp);
 		}
+		fclose(fp);
 	}
 
-	memset(my_ns, 0, sizeof(my_ns));
-	memset(proc_ns, 0, sizeof(proc_ns));
-	if (exists("/proc/self/ns/pid")) {
-		rc = readlink("/proc/self/ns/pid", my_ns, sizeof(my_ns)-1);
-		if (rc <= 0)
-			my_ns[0] = '\0';
-	}
+	if ((rc = readlink("/proc/self/ns/pid", my_ns, sizeof(my_ns)-1)) <= 0)
+		my_ns[0] = '\0';
 
 	while ((entry = readdir(procdir)) != NULL) {
 		if (sscanf(entry->d_name, "%d", &p) != 1)
@@ -166,11 +158,8 @@ rc_find_pids(const char *exec, const char *const *argv, uid_t uid, pid_t pid)
 		if (pid != 0 && pid != p)
 			continue;
 		xasprintf(&buffer, "/proc/%d/ns/pid", p);
-		if (exists(buffer)) {
-			rc = readlink(buffer, proc_ns, sizeof(proc_ns)-1);
-			if (rc <= 0)
-				proc_ns[0] = '\0';
-		}
+		if ((rc = readlink(buffer, proc_ns, sizeof(proc_ns)-1)) <= 0)
+			proc_ns[0] = '\0';
 		free(buffer);
 		if (pid == 0 && strlen(my_ns) && strlen (proc_ns) && strcmp(my_ns, proc_ns))
 			continue;
@@ -190,19 +179,17 @@ rc_find_pids(const char *exec, const char *const *argv, uid_t uid, pid_t pid)
 		/* If this is an OpenVZ host, filter out container processes */
 		if (openvz_host) {
 			xasprintf(&buffer, "/proc/%d/status", p);
-			if (exists(buffer)) {
-				fp = fopen(buffer, "r");
-				free(buffer);
-				if (!fp)
-					continue;
-				while (xgetline(&line, &len, fp) != -1) {
-					if (strncmp(line, "envID:", 6) == 0) {
-						container_pid = !(strncmp(line, "envID:\t0", 8) == 0);
-						break;
-					}
+			fp = fopen(buffer, "r");
+			free(buffer);
+			if (!fp)
+				continue;
+			while (xgetline(&line, &len, fp) != -1) {
+				if (strncmp(line, "envID:", 6) == 0) {
+					container_pid = !(strncmp(line, "envID:\t0", 8) == 0);
+					break;
 				}
-				fclose(fp);
 			}
+			fclose(fp);
 		}
 		if (container_pid)
 			continue;
@@ -329,17 +316,17 @@ rc_find_pids(const char *exec, const char *const *argv, uid_t uid, pid_t pid)
 #endif
 
 static bool
-_match_daemon(const char *path, const char *file, RC_STRINGLIST *match)
+_match_daemon(const char *svcname, const char *instance, RC_STRINGLIST *match)
 {
 	char *line = NULL;
 	size_t len = 0;
-	char *ffile = NULL;
 	FILE *fp;
 	RC_STRING *m;
+	char *daemon;
 
-	xasprintf(&ffile, "%s/%s", path, file);
-	fp = fopen(ffile, "r");
-	free(ffile);
+	xasprintf(&daemon, "%s/%s", svcname, instance);
+	fp = do_fopenat(rc_dirfd(RC_DIR_DAEMONS), daemon, O_RDONLY);
+	free(daemon);
 
 	if (!fp)
 		return false;
@@ -393,7 +380,6 @@ rc_service_daemon_set(const char *service, const char *exec,
     const char *const *argv,
     const char *pidfile, bool started)
 {
-	char *dirpath = NULL;
 	char *file = NULL;
 	int nfiles = 0;
 	char oldfile[PATH_MAX] = { '\0' };
@@ -403,23 +389,22 @@ rc_service_daemon_set(const char *service, const char *exec,
 	RC_STRINGLIST *match, *renamelist;
 	int i = 0;
 	FILE *fp;
+	const char *base = basename_c(service);
 
 	if (!exec && !pidfile) {
 		errno = EINVAL;
 		return false;
 	}
 
-	xasprintf(&dirpath, "%s/daemons/%s", rc_svcdir(), basename_c(service));
-
 	/* Regardless, erase any existing daemon info */
-	if ((dp = opendir(dirpath))) {
+	if ((dp = do_opendirat(rc_dirfd(RC_DIR_DAEMONS), base))) {
 		match = _match_list(exec, argv, pidfile);
 		renamelist = rc_stringlist_new();
 		while ((d = readdir(dp))) {
 			if (d->d_name[0] == '.')
 				continue;
 
-			xasprintf(&file, "%s/%s", dirpath, d->d_name);
+			xasprintf(&file, "%s/daemons/%s/%s", rc_svcdir(), base, d->d_name);
 			if (rc_stringlist_find(renamelist, file)) {
 				free(file);
 				continue;
@@ -428,7 +413,7 @@ rc_service_daemon_set(const char *service, const char *exec,
 			nfiles++;
 
 			if (!*oldfile) {
-				if (_match_daemon(dirpath, d->d_name, match)) {
+				if (_match_daemon(base, d->d_name, match)) {
 					unlink(file);
 					strlcpy(oldfile, file, sizeof(oldfile));
 					nfiles--;
@@ -448,30 +433,29 @@ rc_service_daemon_set(const char *service, const char *exec,
 	}
 
 	/* Now store our daemon info */
-	if (started) {
-		if (mkdir(dirpath, 0755) == 0 || errno == EEXIST) {
-			xasprintf(&file, "%s/%03d", dirpath, nfiles + 1);
-			if ((fp = fopen(file, "w"))) {
-				fprintf(fp, "exec=");
-				if (exec)
-					fprintf(fp, "%s", exec);
-				while (argv && argv[i]) {
-					fprintf(fp, "\nargv_%d=%s", i, argv[i]);
-					i++;
-				}
-				fprintf(fp, "\npidfile=");
-				if (pidfile)
-					fprintf(fp, "%s", pidfile);
-				fprintf(fp, "\n");
-				fclose(fp);
-				retval = true;
+	if (!started)
+		return true;
+
+	if (mkdirat(rc_dirfd(RC_DIR_DAEMONS), base, 0755) == 0 || errno == EEXIST) {
+		xasprintf(&file, "%s/daemons/%s/%03d", rc_svcdir(), base, nfiles + 1);
+		if ((fp = fopen(file, "w"))) {
+			fprintf(fp, "exec=");
+			if (exec)
+				fprintf(fp, "%s", exec);
+			while (argv && argv[i]) {
+				fprintf(fp, "\nargv_%d=%s", i, argv[i]);
+				i++;
 			}
-			free(file);
+			fprintf(fp, "\npidfile=");
+			if (pidfile)
+				fprintf(fp, "%s", pidfile);
+			fprintf(fp, "\n");
+			fclose(fp);
+			retval = true;
 		}
-	} else
-		retval = true;
+		free(file);
+	}
 
-	free(dirpath);
 	return retval;
 }
 
@@ -479,7 +463,7 @@ bool
 rc_service_started_daemon(const char *service,
     const char *exec, const char *const *argv, int indx)
 {
-	char *dirpath = NULL;
+	const char *base = basename_c(service);
 	char *file = NULL;
 	RC_STRINGLIST *match;
 	bool retval = false;
@@ -489,38 +473,31 @@ rc_service_started_daemon(const char *service,
 	if (!service || !exec)
 		return false;
 
-	xasprintf(&dirpath, "%s/daemons/%s", rc_svcdir(), basename_c(service));
 	match = _match_list(exec, argv, NULL);
 
 	if (indx > 0) {
 		xasprintf(&file, "%03d", indx);
-		retval = _match_daemon(dirpath, file, match);
+		retval = _match_daemon(base, file, match);
 		free(file);
-	} else {
-		if ((dp = opendir(dirpath))) {
-			while ((d = readdir(dp))) {
-				if (d->d_name[0] == '.')
-					continue;
-				retval = _match_daemon(dirpath, d->d_name, match);
-				if (retval)
-					break;
-			}
-			closedir(dp);
+	} else if ((dp = do_opendirat(rc_dirfd(RC_DIR_DAEMONS), base))) {
+		while ((d = readdir(dp))) {
+			if (d->d_name[0] == '.')
+				continue;
+			if ((retval = _match_daemon(base, d->d_name, match)))
+				break;
 		}
+		closedir(dp);
 	}
 
 	rc_stringlist_free(match);
-	free(dirpath);
 	return retval;
 }
 
 bool
 rc_service_daemons_crashed(const char *service)
 {
-	char dirpath[PATH_MAX];
 	DIR *dp;
 	struct dirent *d;
-	char *path = dirpath;
 	FILE *fp;
 	char *line = NULL;
 	size_t len = 0;
@@ -541,20 +518,14 @@ rc_service_daemons_crashed(const char *service)
 	char *ch_root;
 	char *spidfile;
 
-	path += snprintf(dirpath, sizeof(dirpath),
-			"%s/daemons/%s", rc_svcdir(), basename_c(service));
-
-	if (!(dp = opendir(dirpath)))
+	if (!(dp = do_opendirat(rc_dirfd(RC_DIR_DAEMONS), basename_c(service))))
 		return false;
 
 	while ((d = readdir(dp))) {
 		if (d->d_name[0] == '.')
 			continue;
 
-		snprintf(path, sizeof(dirpath) - (path - dirpath), "/%s",
-		    d->d_name);
-		fp = fopen(dirpath, "r");
-		if (!fp)
+		if (!(fp = do_fopenat(dirfd(dp), d->d_name, O_RDONLY)))
 			break;
 
 		while (xgetline(&line, &len, fp) != -1) {
@@ -587,9 +558,7 @@ rc_service_daemons_crashed(const char *service)
 		ch_root = rc_service_value_get(basename_c(service), "chroot");
 		spidfile = pidfile;
 		if (ch_root && pidfile) {
-			spidfile = xmalloc(strlen(ch_root) + strlen(pidfile) + 1);
-			strcpy(spidfile, ch_root);
-			strcat(spidfile, pidfile);
+			xasprintf(&spidfile, "%s%s", ch_root, pidfile);
 			free(pidfile);
 			pidfile = spidfile;
 		}
