@@ -145,22 +145,9 @@ make_deptree(void) {
 	return deptree;
 }
 
-RC_DEPTREE *
-rc_deptree_load(void) {
-	char *deptree_cache;
-	RC_DEPTREE *deptree;
-
-	xasprintf(&deptree_cache, "%s/deptree", rc_svcdir());
-	deptree = rc_deptree_load_file(deptree_cache);
-	free(deptree_cache);
-
-	return deptree;
-}
-
-RC_DEPTREE *
-rc_deptree_load_file(const char *deptree_file)
+static RC_DEPTREE *
+deptree_load_file(int dirfd, const char *pathname)
 {
-	FILE *fp;
 	RC_DEPTREE *deptree;
 	RC_DEPINFO *depinfo = NULL;
 	RC_DEPTYPE *deptype = NULL;
@@ -170,8 +157,9 @@ rc_deptree_load_file(const char *deptree_file)
 	char *p;
 	char *e;
 	int i;
+	FILE *fp;
 
-	if (!(fp = fopen(deptree_file, "r")))
+	if (!(fp = do_fopenat(dirfd, pathname, O_RDONLY)))
 		return NULL;
 
 	deptree = make_deptree();
@@ -205,12 +193,24 @@ rc_deptree_load_file(const char *deptree_file)
 			deptype = make_deptype(depinfo, type);
 		rc_stringlist_add(deptype->services, e);
 	}
-	fclose(fp);
 	free(line);
+	fclose(fp);
 
 	return deptree;
 }
 
+RC_DEPTREE *
+rc_deptree_load(void)
+{
+	return deptree_load_file(rc_dirfd(RC_DIR_SVCDIR), "deptree");
+}
+
+RC_DEPTREE *
+rc_deptree_load_file(const char *deptree_file)
+{
+	return deptree_load_file(AT_FDCWD, deptree_file);
+}
+
 static bool
 valid_service(const char *runlevel, const char *service, const char *type)
 {
@@ -293,11 +293,11 @@ get_provided1(const char *runlevel, RC_STRINGLIST *providers,
 /* Work out if a service is provided by another service.
    For example metalog provides logger.
    We need to be able to handle syslogd providing logger too.
-   We do this by checking whats running, then what's starting/stopping,
+   We do this by checking what's running, then what's starting/stopping,
    then what's run in the runlevels and finally alphabetical order.
 
    If there are any bugs in rc-depend, they will probably be here as
-   provided dependancy can change depending on runlevel state.
+   provided dependency can change depending on runlevel state.
    */
 static RC_STRINGLIST *
 get_provided(const RC_DEPINFO *depinfo, const char *runlevel, int options)
@@ -584,18 +584,15 @@ rc_deptree_order(const RC_DEPTREE *deptree, const char *runlevel, int options)
    oldest (or newest) found.
 */
 static bool
-deep_mtime_check(const char *target, bool newer,
-	    time_t *rel, char *file)
+deep_mtime_check(int target_dir, const char *target, bool newer, time_t *rel, char *file)
 {
 	struct stat buf;
 	bool retval = true;
 	DIR *dp;
 	struct dirent *d;
-	char path[PATH_MAX];
-	int serrno = errno;
 
 	/* If target does not exist, return true to mimic shell test */
-	if (stat(target, &buf) != 0)
+	if (fstatat(target_dir, target, &buf, 0) != 0)
 		return true;
 
 	if (newer) {
@@ -616,21 +613,18 @@ deep_mtime_check(const char *target, bool newer,
 		}
 	}
 
-	/* If not a dir then reset errno */
-	if (!(dp = opendir(target))) {
-		errno = serrno;
+	if (!S_ISDIR(buf.st_mode) || !(dp = do_opendirat(target_dir, target)))
 		return retval;
-	}
 
 	/* Check all the entries in the dir */
 	while ((d = readdir(dp))) {
 		if (d->d_name[0] == '.')
 			continue;
-		snprintf(path, sizeof(path), "%s/%s", target, d->d_name);
-		if (!deep_mtime_check(path, newer, rel, file)) {
+		if (!deep_mtime_check(dirfd(dp), d->d_name, newer, rel, file)) {
 			retval = false;
 		}
 	}
+
 	closedir(dp);
 	return retval;
 }
@@ -640,8 +634,7 @@ deep_mtime_check(const char *target, bool newer,
  * the return value arguments are non-null).
  */
 static bool
-mtime_check(const char *source, const char *target, bool newer,
-	    time_t *rel, char *file)
+mtime_check(int dirfd, const char *source, const char *target, bool newer, time_t *rel, char *file)
 {
 	struct stat buf;
 	time_t mtime;
@@ -652,7 +645,7 @@ mtime_check(const char *source, const char *target, bool newer,
 		return false;
 	mtime = buf.st_mtime;
 
-	retval = deep_mtime_check(target,newer,&mtime,file);
+	retval = deep_mtime_check(dirfd, target,newer,&mtime,file);
 	if (rel) {
 		*rel = mtime;
 	}
@@ -660,18 +653,16 @@ mtime_check(const char *source, const char *target, bool newer,
 }
 
 bool
-rc_newer_than(const char *source, const char *target,
-	      time_t *newest, char *file)
+rc_newer_than(const char *source, const char *target, time_t *newest, char *file)
 {
 
-	return mtime_check(source, target, true, newest, file);
+	return mtime_check(AT_FDCWD, source, target, true, newest, file);
 }
 
 bool
-rc_older_than(const char *source, const char *target,
-	      time_t *oldest, char *file)
+rc_older_than(const char *source, const char *target, time_t *oldest, char *file)
 {
-	return mtime_check(source, target, false, oldest, file);
+	return mtime_check(AT_FDCWD, source, target, false, oldest, file);
 }
 
 typedef struct deppair
@@ -690,50 +681,24 @@ static const DEPPAIR deppairs[] = {
 	{ NULL, NULL }
 };
 
-static const char *const depdirs[] =
-{
-	"",
-	"starting",
-	"started",
-	"stopping",
-	"inactive",
-	"wasinactive",
-	"failed",
-	"hotplugged",
-	"daemons",
-	"options",
-	"exclusive",
-	"scheduled",
-	"tmp",
-	NULL
-};
-
 bool
 rc_deptree_update_needed(time_t *newest, char *file)
 {
 	bool newer = false;
+	const int *dirfds;
 	RC_STRINGLIST *config;
 	RC_STRING *s;
-	int i;
 	struct stat buf;
 	time_t mtime;
-	char *path;
-	char *deptree_cache, *depconfig;
-	const char *service_dir = rc_svcdir();
-
-	/* Create base directories if needed */
-	for (i = 0; depdirs[i]; i++) {
-		xasprintf(&path, "%s/%s", service_dir, depdirs[i]);
-		if (mkdir(path, 0755) != 0 && errno != EEXIST)
-			fprintf(stderr, "mkdir `%s': %s\n", depdirs[i], strerror(errno));
-		free(path);
-	}
 
 	/* Quick test to see if anything we use has changed and we have
 	 * data in our deptree. */
+	if (mkdir(rc_svcdir(), 0755) != 0 && errno != EEXIST)
+		fprintf(stderr, "mkdir '%s': %s\n", rc_svcdir(), strerror(errno));
 
-	xasprintf(&deptree_cache, "%s/deptree", service_dir);
-	if (stat(deptree_cache, &buf) == 0) {
+	clear_dirfds(); /* clear our cached dirfds as a sanity check */
+
+	if (fstatat(rc_dirfd(RC_DIR_SVCDIR), "deptree", &buf, 0) == 0) {
 		mtime = buf.st_mtime;
 	} else {
 		/* No previous cache found.
@@ -743,30 +708,25 @@ rc_deptree_update_needed(time_t *newest, char *file)
 		newer = true;
 		mtime = time(NULL);
 	}
-	free(deptree_cache);
 
-	for (const char * const *dirs = rc_scriptdirs(); *dirs; dirs++) {
-		static const char *subdirs[] = { "init.d", "conf.d", NULL };
-		for (const char **subdir = subdirs; *subdir; subdir++) {
-			xasprintf(&path, "%s/%s", *dirs, *subdir);
-			newer |= !deep_mtime_check(path, true, &mtime, file);
-			free(path);
-		}
+	for (size_t i = 0, count = rc_scriptdirfds(&dirfds); i < count; i++) {
+		newer |= !deep_mtime_check(dirfds[i], "init.d", true, &mtime, file);
+		newer |= !deep_mtime_check(dirfds[i], "conf.d", true, &mtime, file);
 	}
 
-	xasprintf(&path, "%s/rc.conf", rc_sysconfdir());
-	newer |= !deep_mtime_check(path, true, &mtime, file);
-	free(path);
+	newer |= !deep_mtime_check(rc_dirfd(RC_DIR_SYSCONF), "rc.conf", true, &mtime, file);
+	newer |= !deep_mtime_check(rc_dirfd(RC_DIR_SYSCONF), "rc.conf.d", true, &mtime, file);
+	if (rc_is_user()) {
+		newer |= !deep_mtime_check(rc_dirfd(RC_DIR_USRCONF), "rc.conf", true, &mtime, file);
+		newer |= !deep_mtime_check(rc_dirfd(RC_DIR_USRCONF), "rc.conf.d", true, &mtime, file);
+	}
 
 	/* Some init scripts dependencies change depending on config files
 	 * outside of baselayout, like syslog-ng, so we check those too. */
-	xasprintf(&depconfig, "%s/depconfig", service_dir);
-	config = rc_config_list(depconfig);
-	TAILQ_FOREACH(s, config, entries) {
-		newer |= !deep_mtime_check(s->value, true, &mtime, file);
-	}
+	config = config_list(rc_dirfd(RC_DIR_SVCDIR), "depconfig");
+	TAILQ_FOREACH(s, config, entries)
+		newer |= !deep_mtime_check(AT_FDCWD, s->value, true, &mtime, file);
 	rc_stringlist_free(config);
-	free(depconfig);
 
 	/* Return newest file time, if requested */
 	if ((newer) && (newest != NULL)) {
@@ -777,21 +737,19 @@ rc_deptree_update_needed(time_t *newest, char *file)
 }
 
 static void
-setup_environment(void) {
-	char *scriptdirs, *env;
-	size_t env_size = 0;
+setup_environment(void)
+{
 	struct utsname uts;
-
-	for (const char * const *dirs = rc_scriptdirs(); *dirs; dirs++)
-		env_size += strlen(*dirs) + sizeof(' ');
-
-	env = scriptdirs = xmalloc(env_size);
+	size_t env_size;
+	char *env;
+	FILE *mem = xopen_memstream(&env, &env_size);
 
 	for (const char * const *dirs = rc_scriptdirs(); *dirs; dirs++) {
-		int len = snprintf(scriptdirs, env_size, "%s ", *dirs);
-		scriptdirs += len;
-		env_size -= len;
+		fputs(*dirs, mem);
+		if (dirs[1])
+			fputc(' ', mem);
 	}
+	xclose_memstream(mem);
 
 	setenv("RC_SCRIPTDIRS", env, 1);
 	free(env);
@@ -829,7 +787,6 @@ rc_deptree_update(void)
 	char *line = NULL;
 	size_t size;
 	char *depend, *depends, *service, *type;
-	char *deptree_cache, *depconfig;
 	size_t i, l;
 	bool retval = true;
 	const char *sys = rc_sys();
@@ -1082,8 +1039,7 @@ rc_deptree_update(void)
 	   This works and should be entirely shell parseable provided that depend
 	   names don't have any non shell variable characters in
 	   */
-	xasprintf(&deptree_cache, "%s/deptree", rc_svcdir());
-	if ((fp = fopen(deptree_cache, "w"))) {
+	if ((fp = do_fopenat(rc_dirfd(RC_DIR_SVCDIR), "deptree", O_WRONLY | O_CREAT | O_TRUNC))) {
 		i = 0;
 		TAILQ_FOREACH(depinfo, deptree, entries) {
 			fprintf(fp, "depinfo_%zu_service='%s'\n", i, depinfo->service);
@@ -1098,26 +1054,23 @@ rc_deptree_update(void)
 		}
 		fclose(fp);
 	} else {
-		fprintf(stderr, "fopen '%s': %s\n", deptree_cache, strerror(errno));
+		fprintf(stderr, "fopen '%s/deptree': %s\n", rc_svcdir(), strerror(errno));
 		retval = false;
 	}
-	free(deptree_cache);
 
 	/* Save our external config files to disk */
-	xasprintf(&depconfig, "%s/depconfig", rc_svcdir());
 	if (TAILQ_FIRST(config)) {
-		if ((fp = fopen(depconfig, "w"))) {
+		if ((fp = do_fopenat(rc_dirfd(RC_DIR_SVCDIR), "depconfig", O_WRONLY | O_CREAT | O_TRUNC))) {
 			TAILQ_FOREACH(s, config, entries)
 				fprintf(fp, "%s\n", s->value);
 			fclose(fp);
 		} else {
-			fprintf(stderr, "fopen '%s': %s\n", depconfig, strerror(errno));
+			fprintf(stderr, "fopen '%s/depconfig': %s\n", rc_svcdir(), strerror(errno));
 			retval = false;
 		}
 	} else {
-		unlink(depconfig);
+		unlinkat(rc_dirfd(RC_DIR_SVCDIR), "depconfig", 0);
 	}
-	free(depconfig);
 
 	rc_stringlist_free(config);
 	rc_deptree_free(deptree);
