@@ -15,6 +15,7 @@
  *    except according to the terms contained in the LICENSE file.
  */
 
+#include <assert.h>
 #include <helpers.h>
 #include <ctype.h>
 #include <dirent.h>
@@ -45,122 +46,131 @@
 
 /* File stream used for plugins to write environ vars to */
 FILE *rc_environ_fd = NULL;
+static int dirfds[RC_DIR_MAX];
 
-typedef struct rc_service_state_name {
-	RC_SERVICE state;
-	const char *name;
-} rc_service_state_name_t;
+int rc_dirfd(enum rc_dir dir) {
+	int flags = O_RDONLY | O_CLOEXEC | O_DIRECTORY;
+	int dirfd;
 
-/* We MUST list the states below 0x10 first
- * The rest can be in any order */
-static const rc_service_state_name_t rc_service_state_names[] = {
-	{ RC_SERVICE_STARTED,     "started" },
-	{ RC_SERVICE_STOPPED,     "stopped" },
-	{ RC_SERVICE_STARTING,    "starting" },
-	{ RC_SERVICE_STOPPING,    "stopping" },
-	{ RC_SERVICE_INACTIVE,    "inactive" },
-	{ RC_SERVICE_WASINACTIVE, "wasinactive" },
-	{ RC_SERVICE_HOTPLUGGED,  "hotplugged" },
-	{ RC_SERVICE_FAILED,      "failed" },
-	{ RC_SERVICE_SCHEDULED,   "scheduled"},
-	{ RC_SERVICE_CRASHED,     "crashed"},
-	{ 0, NULL}
-};
+	if (dir >= RC_DIR_MAX)
+		return -1;
+	if (dirfds[dir] <= 0) {
+		switch (dir) {
+		case RC_DIR_INVALID:
+			return -1;
+		case RC_DIR_SYSCONF:
+			return dirfds[dir] = open(rc_sysconfdir(), flags);
+		case RC_DIR_USRCONF:
+			return dirfds[dir] = open(rc_usrconfdir(), flags);
+		case RC_DIR_RUNLEVEL:
+			return dirfds[dir] = open(rc_runleveldir(), flags);
+		case RC_DIR_SVCDIR:
+			return dirfds[dir] = open(rc_svcdir(), flags);
+		default:
+			assert(dirnames[dir]);
+			dirfd = rc_dirfd(RC_DIR_SVCDIR);
+			if ((dirfds[dir] = openat(dirfd, dirnames[dir], flags)) == -1 && errno == ENOENT) {
+				if (mkdirat(dirfd, dirnames[dir], 0755) == -1 && errno != EEXIST)
+					return -1;
+				dirfds[dir] = openat(dirfd, dirnames[dir], flags);
+			}
+			break;
+		}
+	}
+	return dirfds[dir];
+}
+
+void clear_dirfds(void) {
+	for (size_t i = 0; i < RC_DIR_MAX; i++) {
+		/* dirfd is zero-initialized, though zero is a valid
+		 * FD, dirfd should never contain FDs which are <= 2. */
+		if (dirfds[i] <= 0)
+			continue;
+
+		close(dirfds[i]);
+		dirfds[i] = -1;
+	}
+}
 
 #define LS_INITD	0x01
 #define LS_DIR		0x02
 static RC_STRINGLIST *
-ls_dir(const char *dir, int options)
+ls_dir(int targetfd, const char *dir, int options)
 {
 	DIR *dp;
 	struct dirent *d;
 	RC_STRINGLIST *list = NULL;
 	struct stat buf;
 	size_t l;
-	char *file;
-	int r;
 
 	list = rc_stringlist_new();
-	if ((dp = opendir(dir)) == NULL)
-		return list;
-	while (((d = readdir(dp)) != NULL)) {
-		if (d->d_name[0] != '.') {
-			if (options & LS_INITD) {
-				/* Check that our file really exists.
-				 * This is important as a service maybe in a
-				 * runlevel, but could have been removed. */
-				xasprintf(&file, "%s/%s", dir, d->d_name);
-				r = stat(file, &buf);
-				free(file);
-				if (r != 0)
-					continue;
 
-				/* .sh files are not init scripts */
-				l = strlen(d->d_name);
-				if (l > 2 && d->d_name[l - 3] == '.' &&
-				    d->d_name[l - 2] == 's' &&
-				    d->d_name[l - 1] == 'h')
-					continue;
-			}
-			if (options & LS_DIR) {
-				xasprintf(&file, "%s/%s", dir, d->d_name);
-				r = stat(file, &buf);
-				free(file);
-				if (r != 0 || !S_ISDIR(buf.st_mode))
-					continue;
-			}
-			rc_stringlist_add(list, d->d_name);
+	if (!(dp = do_opendirat(targetfd, dir)))
+		return list;
+
+	while (((d = readdir(dp)) != NULL)) {
+		if (d->d_name[0] == '.')
+			continue;
+		if (options & LS_INITD) {
+			/* Check that our file really exists.
+			 * This is important as a service maybe in a
+			 * runlevel, but could have been removed. */
+			if (fstatat(dirfd(dp), d->d_name, &buf, 0) != 0)
+				continue;
+
+			/* .sh files are not init scripts */
+			l = strlen(d->d_name);
+			if (l > 2 && d->d_name[l - 3] == '.' &&
+					d->d_name[l - 2] == 's' &&
+					d->d_name[l - 1] == 'h')
+				continue;
 		}
+		if (options & LS_DIR) {
+			if (fstatat(dirfd(dp), d->d_name, &buf, 0) != 0 || !S_ISDIR(buf.st_mode))
+				continue;
+		}
+		rc_stringlist_add(list, d->d_name);
 	}
 	closedir(dp);
 	return list;
 }
 
 static bool
-rm_dir(const char *pathname, bool top)
+rm_dir(int targetfd, const char *pathname, bool top)
 {
-	DIR *dp;
-	struct dirent *d;
-	char *file = NULL;
-	struct stat s;
 	bool retval = true;
+	struct dirent *d;
+	struct stat s;
+	DIR *dp;
 
-	if ((dp = opendir(pathname)) == NULL)
+	if (!(dp = do_opendirat(targetfd, pathname)))
 		return false;
 
-	errno = 0;
-	while (((d = readdir(dp)) != NULL) && errno == 0) {
-		if (strcmp(d->d_name, ".") != 0 &&
-		    strcmp(d->d_name, "..") != 0)
-		{
-			xasprintf(&file, "%s/%s", pathname, d->d_name);
-			if (stat(file, &s) != 0) {
+	while ((d = readdir(dp))) {
+		if (strcmp(d->d_name, ".") == 0 || strcmp(d->d_name, "..") == 0)
+			continue;
+		if (fstatat(dirfd(dp), d->d_name, &s, 0) != 0) {
+			retval = false;
+			break;
+		}
+		if (S_ISDIR(s.st_mode)) {
+			if (!rm_dir(dirfd(dp), d->d_name, true)) {
 				retval = false;
 				break;
 			}
-			if (S_ISDIR(s.st_mode)) {
-				if (!rm_dir(file, true))
-				{
-					retval = false;
-					break;
-				}
-			} else {
-				if (unlink(file)) {
-					retval = false;
-					break;
-				}
+		} else {
+			if (unlinkat(dirfd(dp), d->d_name, 0) != 0) {
+				retval = false;
+				break;
 			}
-			free(file);
-			file = NULL;
 		}
 	}
-	free(file);
 	closedir(dp);
 
 	if (!retval)
 		return false;
 
-	if (top && rmdir(pathname) != 0)
+	if (top && unlinkat(targetfd, pathname, AT_REMOVEDIR) != 0)
 		return false;
 
 	return true;
@@ -183,11 +193,10 @@ file_regex(const char *file, const char *regex)
 		return false;
 
 	if ((result = regcomp(&re, regex, REG_EXTENDED | REG_NOSUB)) != 0) {
+		char buf[BUFSIZ];
 		fclose(fp);
-		line = xmalloc(sizeof(char) * BUFSIZ);
-		regerror(result, &re, line, BUFSIZ);
-		fprintf(stderr, "file_regex: %s", line);
-		free(line);
+		regerror(result, &re, buf, sizeof(buf));
+		fprintf(stderr, "file_regex: %s", buf);
 		return false;
 	}
 
@@ -252,7 +261,7 @@ detect_container(const char *systype RC_UNUSED)
 		if (jailed == 1)
 			return RC_SYS_JAIL;
 
-	if (exists("/var/run/.containerenv"))
+	if (access("/var/run/.containerenv", F_OK) == 0)
 		return RC_SYS_PODMAN;
 #endif
 
@@ -276,13 +285,15 @@ detect_container(const char *systype RC_UNUSED)
 				return RC_SYS_DOCKER;
 		if (strcmp(systype, RC_SYS_PODMAN) == 0)
 			return RC_SYS_PODMAN;
+		if (strcmp(systype, RC_SYS_WSL) == 0)
+			return RC_SYS_WSL;
 	}
 	if (file_regex("/proc/cpuinfo", "UML"))
 		return RC_SYS_UML;
 	else if (file_regex("/proc/self/status",
 		"(s_context|VxID):[[:space:]]*[1-9]"))
 		return RC_SYS_VSERVER;
-	else if (exists("/proc/vz/veinfo") && !exists("/proc/vz/version"))
+	else if (access("/proc/vz/veinfo", F_OK) == 0 && access("/proc/vz/version", F_OK) != 0)
 		return RC_SYS_OPENVZ;
 	else if (file_regex("/proc/self/status",
 		"envID:[[:space:]]*[1-9]"))
@@ -293,13 +304,19 @@ detect_container(const char *systype RC_UNUSED)
 		return RC_SYS_RKT;
 	else if (file_regex("/proc/1/environ", "container=systemd-nspawn"))
 		return RC_SYS_SYSTEMD_NSPAWN;
-	else if (exists("/run/.containerenv"))
+	else if (access("/run/.containerenv", F_OK) == 0)
 		return RC_SYS_PODMAN;
-	else if (exists("/.dockerenv"))
+	else if (access("/.dockerenv", F_OK) == 0)
 		return RC_SYS_DOCKER;
 	/* old test, I'm not sure when this was valid. */
 	else if (file_regex("/proc/1/environ", "container=docker"))
 		return RC_SYS_DOCKER;
+	else if (file_regex("/proc/sys/kernel/osrelease", "microsoft"))
+		return RC_SYS_WSL;
+	else if (getenv("WSL_DISTRO_NAME"))
+		return RC_SYS_WSL;
+	else if (access("/proc/sys/fs/binfmt_misc/WSLInterop", F_OK) == 0)
+		return RC_SYS_WSL;
 #endif
 
 	return NULL;
@@ -317,9 +334,9 @@ detect_vm(const char *systype RC_UNUSED)
 		if (strcmp(systype, RC_SYS_XENU) == 0)
 			return RC_SYS_XENU;
 	}
-	if (exists("/kern/xen/privcmd"))
+	if (access("/kern/xen/privcmd", F_OK) == 0)
 		return RC_SYS_XEN0;
-	if (exists("/kern/xen"))
+	if (access("/kern/xen", F_OK) == 0)
 		return RC_SYS_XENU;
 #endif
 
@@ -332,7 +349,7 @@ detect_vm(const char *systype RC_UNUSED)
 		if (strcmp(systype, RC_SYS_XENU) == 0)
 			return RC_SYS_XENU;
 	}
-	if (exists("/proc/xen")) {
+	if (access("/proc/xen", F_OK) == 0) {
 		if (file_regex("/proc/xen/capabilities", "control_d"))
 			return RC_SYS_XEN0;
 		return RC_SYS_XENU;
@@ -369,13 +386,21 @@ rc_parse_service_state(RC_SERVICE state)
 	return NULL;
 }
 
+static int
+rc_parse_service_state_dirfd(RC_SERVICE state)
+{
+	for (size_t i = 0; rc_service_state_names[i].name; i++)
+		if (rc_service_state_names[i].state == state)
+			return rc_service_state_names[i].dir;
+	return -1;
+}
+
 /* Returns a list of all the chained runlevels used by the
  * specified runlevel in dependency order, including the
  * specified runlevel. */
 static void
 get_runlevel_chain(const char *runlevel, RC_STRINGLIST *level_list, RC_STRINGLIST *ancestor_list)
 {
-	char *path;
 	RC_STRINGLIST *dirs;
 	RC_STRING *d, *parent;
 	const char *nextlevel;
@@ -400,8 +425,7 @@ get_runlevel_chain(const char *runlevel, RC_STRINGLIST *level_list, RC_STRINGLIS
 	 * We can now do exactly the above procedure for our chained
 	 * runlevels.
 	 */
-	xasprintf(&path, "%s/%s", rc_runleveldir(), runlevel);
-	dirs = ls_dir(path, LS_DIR);
+	dirs = ls_dir(rc_dirfd(RC_DIR_RUNLEVEL), runlevel, LS_DIR);
 	TAILQ_FOREACH(d, dirs, entries) {
 		nextlevel = d->value;
 
@@ -423,47 +447,33 @@ get_runlevel_chain(const char *runlevel, RC_STRINGLIST *level_list, RC_STRINGLIS
 		rc_stringlist_delete(ancestor_list, nextlevel);
 	}
 	rc_stringlist_free(dirs);
-	free(path);
-}
-
-static bool
-svcdir_subpath_exists(const char *subdir)
-{
-	char *path;
-	bool found;
-	xasprintf(&path, "%s/%s", rc_svcdir(), subdir);
-	found = exists(path);
-	free(path);
-	return found;
 }
 
 bool
 rc_runlevel_starting(void)
 {
-	return svcdir_subpath_exists("rc.starting");
+	return faccessat(rc_dirfd(RC_DIR_SVCDIR), "rc.starting", F_OK, 0) == 0;
 }
 
 bool
 rc_runlevel_stopping(void)
 {
-	return svcdir_subpath_exists("rc.stopping");
+	return faccessat(rc_dirfd(RC_DIR_SVCDIR), "rc.stopping", F_OK, 0) == 0;
 }
 
 RC_STRINGLIST *rc_runlevel_list(void)
 {
-	return ls_dir(rc_runleveldir(), LS_DIR);
+	return ls_dir(AT_FDCWD, rc_runleveldir(), LS_DIR);
 }
 
 char *
 rc_runlevel_get(void)
 {
 	FILE *fp;
-	char *softlevel;
 	char *runlevel = NULL;
 	size_t i;
 
-	xasprintf(&softlevel, "%s/softlevel", rc_svcdir());
-	if ((fp = fopen(softlevel, "r"))) {
+	if ((fp = do_fopenat(rc_dirfd(RC_DIR_SVCDIR), "softlevel", O_RDONLY))) {
 		runlevel = xmalloc(sizeof(char) * PATH_MAX);
 		if (fgets(runlevel, PATH_MAX, fp)) {
 			i = strlen(runlevel) - 1;
@@ -473,7 +483,6 @@ rc_runlevel_get(void)
 			*runlevel = '\0';
 		fclose(fp);
 	}
-	free(softlevel);
 
 	if (!runlevel || !*runlevel) {
 		free(runlevel);
@@ -486,35 +495,24 @@ rc_runlevel_get(void)
 bool
 rc_runlevel_set(const char *runlevel)
 {
-	char *softlevel;
-	bool ret = false;
 	FILE *fp;
 
-	xasprintf(&softlevel, "%s/softlevel", rc_svcdir());
-	if ((fp = fopen(softlevel, "w"))) {
-		fprintf(fp, "%s", runlevel);
-		fclose(fp);
-		ret = true;
-	}
+	if (!(fp = do_fopenat(rc_dirfd(RC_DIR_SVCDIR), "softlevel", O_WRONLY | O_CREAT | O_TRUNC)))
+		return false;
 
-	free(softlevel);
-	return ret;
+	fputs(runlevel, fp);
+	fclose(fp);
+	return true;
 }
 
 bool
 rc_runlevel_exists(const char *runlevel)
 {
-	char *path;
 	struct stat buf;
-	int r;
 
-	if (!runlevel || strcmp(runlevel, "") == 0 || strcmp(runlevel, ".") == 0 ||
-		strcmp(runlevel, "..") == 0)
+	if (!runlevel || strcmp(runlevel, "") == 0 || strcmp(runlevel, ".") == 0 || strcmp(runlevel, "..") == 0)
 		return false;
-	xasprintf(&path, "%s/%s", rc_runleveldir(), runlevel);
-	r = stat(path, &buf);
-	free(path);
-	if (r == 0 && S_ISDIR(buf.st_mode))
+	if (fstatat(rc_dirfd(RC_DIR_RUNLEVEL), runlevel, &buf, 0) == 0 && S_ISDIR(buf.st_mode))
 		return true;
 	return false;
 }
@@ -559,7 +557,21 @@ rc_runlevel_stacks(const char *runlevel)
 	return stack;
 }
 
-static const char * const scriptdirs[] = {
+enum scriptdirs_entries {
+	SCRIPTDIR_USR,
+#ifdef RC_LOCAL_PREFIX
+	SCRIPTDIR_LOCAL,
+#endif
+	SCRIPTDIR_SYS,
+#ifdef RC_PKG_PREFIX
+	SCRIPTDIR_PKG,
+#endif
+	SCRIPTDIR_SVC,
+	SCRIPTDIR_MAX,
+	SCRIPTDIR_CAP
+};
+
+static const char * const scriptdirs[SCRIPTDIR_CAP] = {
 #ifdef RC_LOCAL_PREFIX
 	RC_LOCAL_PREFIX "/etc",
 #endif
@@ -567,42 +579,150 @@ static const char * const scriptdirs[] = {
 #ifdef RC_PKG_PREFIX
 	RC_PKG_PREFIX "/etc",
 #endif
-	NULL
+	RC_SVCDIR
 };
 
+static struct {
+	bool set;
+	char *svcdir;
+	char *usrconfdir;
+	char *runleveldir;
+	const char *scriptdirs[ARRAY_SIZE(scriptdirs)];
+} rc_dirs = {
+	.scriptdirs = {
+#ifdef RC_LOCAL_PREFIX
+		[SCRIPTDIR_LOCAL] = RC_LOCAL_PREFIX "/etc/user",
+#endif
+		[SCRIPTDIR_SYS] = RC_SYSCONFDIR "/user",
+#ifdef RC_PKG_PREFIX
+		[SCRIPTDIR_PKG] = RC_PKG_PREFIX "/etc/user",
+#endif
+	}
+};
+
+static void
+free_rc_dirs(void)
+{
+	free(rc_dirs.runleveldir);
+	rc_dirs.runleveldir = NULL;
+	free(rc_dirs.svcdir);
+	rc_dirs.svcdir = NULL;
+	rc_dirs.scriptdirs[0] = NULL;
+}
+
+static bool is_user = false;
+
+bool
+rc_is_user(void)
+{
+	return is_user;
+}
+
+void
+rc_set_user(void)
+{
+	char *env;
+	if (is_user)
+		return;
+
+	is_user = true;
+	rc_dirs.set = true;
+	setenv("RC_USER_SERVICES", "yes", true);
+
+	if ((env = getenv("XDG_CONFIG_HOME"))) {
+		xasprintf(&rc_dirs.usrconfdir, "%s/rc", env);
+	} else if ((env = getenv("HOME"))) {
+		xasprintf(&rc_dirs.usrconfdir, "%s/.config/rc", env);
+	} else {
+		fprintf(stderr, "XDG_CONFIG_HOME and HOME unset.\n");
+		exit(EXIT_FAILURE);
+	}
+
+	xasprintf(&rc_dirs.runleveldir, "%s/runlevels", rc_dirs.usrconfdir);
+
+	if (!(env = getenv("XDG_RUNTIME_DIR"))) {
+		/* FIXME: fallback to something else? */
+		fprintf(stderr, "XDG_RUNTIME_DIR unset.\n");
+		exit(EXIT_FAILURE);
+	}
+
+	xasprintf(&rc_dirs.svcdir, "%s/openrc", env);
+	atexit(free_rc_dirs);
+
+	rc_dirs.scriptdirs[SCRIPTDIR_USR] = rc_dirs.usrconfdir;
+	rc_dirs.scriptdirs[SCRIPTDIR_SVC] = rc_dirs.svcdir;
+
+	clear_dirfds();
+}
+
 const char * const *
 rc_scriptdirs(void)
 {
+	if (rc_dirs.set)
+		return rc_dirs.scriptdirs;
 	return scriptdirs;
 }
 
+size_t
+rc_scriptdirfds(const int **fds)
+{
+	static int scriptdirfds[SCRIPTDIR_MAX];
+	static size_t len = 0;
+
+	if (len == 0) {
+		const char * const *current_scriptdirs = rc_scriptdirs();
+		for (size_t i = 0; current_scriptdirs[i]; i++) {
+			int fd = open(current_scriptdirs[i], O_RDONLY | O_DIRECTORY | O_CLOEXEC);
+
+			if (fd == -1)
+				continue;
+
+			scriptdirfds[len++] = fd;
+		}
+	}
+
+	*fds = scriptdirfds;
+	return len;
+}
+
 const char *
 rc_sysconfdir(void)
 {
 	return RC_SYSCONFDIR;
 }
 
+const char *
+rc_usrconfdir(void)
+{
+	if (rc_dirs.set)
+		return rc_dirs.usrconfdir;
+
+	return NULL;
+}
+
 const char *
 rc_runleveldir(void)
 {
+	if (rc_dirs.set)
+		return rc_dirs.runleveldir;
 	return RC_RUNLEVELDIR;
 }
 
 const char *
 rc_svcdir(void)
 {
-	return RC_SVCDIR;
+	return rc_dirs.set ? rc_dirs.svcdir : RC_SVCDIR;
 }
 
 static ssize_t
-safe_readlink(const char *path, char **buf, size_t bufsiz)
+safe_readlink(int targetfd, const char *path, char **buf, size_t bufsiz)
 {
 	char *buffer;
 	ssize_t r;
 
 	for (;; bufsiz += PATH_MAX) {
 		buffer = xmalloc(bufsiz);
-		r = readlink(path, buffer, bufsiz);
+		r = readlinkat(targetfd, path, buffer, bufsiz);
 		if (r < 0) {
 			free(buffer);
 			return -errno;
@@ -623,7 +743,6 @@ char *
 rc_service_resolve(const char *service)
 {
 	char *buffer;
-	char *file = NULL;
 	struct stat buf;
 
 	if (!service)
@@ -633,29 +752,21 @@ rc_service_resolve(const char *service)
 		return xstrdup(service);
 
 	/* First check started services */
-	xasprintf(&file, "%s/%s/%s", rc_svcdir(), "started", service);
-	if (lstat(file, &buf) || !S_ISLNK(buf.st_mode)) {
-		free(file);
-		xasprintf(&file, "%s/%s/%s", rc_svcdir(), "inactive", service);
-		if (lstat(file, &buf) || !S_ISLNK(buf.st_mode)) {
-			free(file);
-			file = NULL;
-		}
-	}
-
-	if (file != NULL && safe_readlink(file, &buffer, buf.st_size + 1) >= 0) {
-		free(file);
-		return buffer;
-	}
+	if (fstatat(rc_dirfd(RC_DIR_STARTED), service, &buf, AT_SYMLINK_NOFOLLOW) == 0 && S_ISLNK(buf.st_mode))
+		if (safe_readlink(rc_dirfd(RC_DIR_STARTED), service, &buffer, buf.st_size + 1) >= 0)
+			return buffer;
+	if (fstatat(rc_dirfd(RC_DIR_INACTIVE), service, &buf, AT_SYMLINK_NOFOLLOW) == 0 && S_ISLNK(buf.st_mode))
+		if (safe_readlink(rc_dirfd(RC_DIR_INACTIVE), service, &buffer, buf.st_size + 1) >= 0)
+			return buffer;
 
 	for (const char * const *dirs = rc_scriptdirs(); *dirs; dirs++) {
-		free(file);
+		char *file = NULL;
 		xasprintf(&file, "%s/init.d/%s", *dirs, service);
 		if (stat(file, &buf) == 0)
 			return file;
+		free(file);
 	}
 
-	free(file);
 	return NULL;
 }
 
@@ -676,8 +787,8 @@ rc_service_exists(const char *service)
 
 	/* .sh files are not init scripts */
 	if (len > 2 && service[len - 3] == '.' &&
-	    service[len - 2] == 's' &&
-	    service[len - 1] == 'h') {
+			service[len - 2] == 's' &&
+			service[len - 1] == 'h') {
 		errno = EINVAL;
 		return false;
 	}
@@ -711,14 +822,11 @@ rc_service_extra_commands(const char *service)
 	char *token;
 	char *p;
 	FILE *fp;
-	size_t l;
 
 	if (!(svc = rc_service_resolve(service)))
 		return NULL;
 
-	l = strlen(OPTSTR) + strlen(svc) + 1;
-	cmd = xmalloc(sizeof(char) * l);
-	snprintf(cmd, l, OPTSTR, svc);
+	xasprintf(&cmd, OPTSTR, svc);
 	free(svc);
 
 	if (!(fp = popen(cmd, "r"))) {
@@ -750,7 +858,6 @@ rc_service_description(const char *service, const char *option)
 	char *desc = NULL;
 	size_t size = 0;
 	FILE *fp;
-	size_t l;
 
 	if (!(svc = rc_service_resolve(service)))
 		return NULL;
@@ -758,9 +865,7 @@ rc_service_description(const char *service, const char *option)
 	if (!option)
 		option = "";
 
-	l = strlen(DESCSTR) + strlen(svc) + strlen(option) + 2;
-	cmd = xmalloc(sizeof(char) * l);
-	snprintf(cmd, l, DESCSTR, svc, *option ? "_" : "", option);
+	xasprintf(&cmd, DESCSTR, svc, *option ? "_" : "", option);
 	free(svc);
 	if ((fp = popen(cmd, "r"))) {
 		if (xgetline(&desc, &size, fp) == -1) {
@@ -780,7 +885,7 @@ rc_service_in_runlevel(const char *service, const char *runlevel)
 	bool r;
 
 	xasprintf(&file, "%s/%s/%s", rc_runleveldir(), runlevel, basename_c(service));
-	r = exists(file);
+	r = access(file, F_OK) == 0;
 	free(file);
 	return r;
 }
@@ -788,16 +893,10 @@ rc_service_in_runlevel(const char *service, const char *runlevel)
 bool
 rc_service_mark(const char *service, const RC_SERVICE state)
 {
-	char *file, *was;
 	int i = 0;
-	int skip_state = -1;
 	const char *base;
 	char *init = rc_service_resolve(service);
-	const char *svcdir = rc_svcdir();
 	bool skip_wasinactive = false;
-	int s;
-	RC_STRINGLIST *dirs;
-	RC_STRING *dir;
 	int serrno;
 
 	if (!init)
@@ -805,21 +904,17 @@ rc_service_mark(const char *service, const RC_SERVICE state)
 
 	base = basename_c(service);
 	if (state != RC_SERVICE_STOPPED) {
-		if (!exists(init)) {
+		int state_dirfd = rc_dirfd(rc_parse_service_state_dirfd(state));
+		if (access(init, F_OK) != 0) {
 			free(init);
 			return false;
 		}
 
-		xasprintf(&file, "%s/%s/%s", svcdir, rc_parse_service_state(state), base);
-		if (exists(file))
-			unlink(file);
-		i = symlink(init, file);
-		free(file);
-		if (i != 0) {
+		unlinkat(state_dirfd, base, 0);
+		if ((symlinkat(init, state_dirfd, base)) != 0) {
 			free(init);
 			return false;
 		}
-		skip_state = state;
 	}
 
 	if (state == RC_SERVICE_HOTPLUGGED || state == RC_SERVICE_FAILED) {
@@ -829,82 +924,68 @@ rc_service_mark(const char *service, const RC_SERVICE state)
 
 	/* Remove any old states now */
 	for (i = 0; rc_service_state_names[i].name; i++) {
-		s = rc_service_state_names[i].state;
+		int state_dirfd = rc_dirfd(rc_service_state_names[i].dir);
+		RC_SERVICE s = rc_service_state_names[i].state;
 
-		if ((s != skip_state &&
-			s != RC_SERVICE_STOPPED &&
-			s != RC_SERVICE_HOTPLUGGED &&
-			s != RC_SERVICE_SCHEDULED) &&
-		    (!skip_wasinactive || s != RC_SERVICE_WASINACTIVE))
-		{
-			xasprintf(&file, "%s/%s/%s", svcdir, rc_service_state_names[i].name, base);
-			if (exists(file)) {
-				if ((state == RC_SERVICE_STARTING ||
-					state == RC_SERVICE_STOPPING) &&
-					s == RC_SERVICE_INACTIVE)
-				{
-					xasprintf(&was, "%s/%s/%s", svcdir,
-						rc_parse_service_state(RC_SERVICE_WASINACTIVE), base);
-					i = symlink(init, was);
-					free(was);
-					if (i == -1) {
-						free(file);
-						free(init);
-						return false;
-					}
-					skip_wasinactive = true;
-				}
-				if (unlink(file) == -1) {
-					free(file);
+		if (s == state)
+			continue;
+
+		switch (s) {
+		case RC_SERVICE_STOPPED:
+		case RC_SERVICE_HOTPLUGGED:
+		case RC_SERVICE_SCHEDULED:
+			continue;
+		case RC_SERVICE_WASINACTIVE:
+			if (skip_wasinactive)
+				continue;
+			/* fall-through */
+		default:
+			if (faccessat(state_dirfd, base, F_OK, 0) != 0)
+				continue;
+			if ((state == RC_SERVICE_STARTING || state == RC_SERVICE_STOPPING) && s == RC_SERVICE_INACTIVE) {
+				if (symlinkat(init, rc_dirfd(RC_DIR_WASINACTIVE), base) == -1) {
 					free(init);
 					return false;
 				}
+				skip_wasinactive = true;
+			}
+			if (unlinkat(state_dirfd, base, 0) == -1) {
+				free(init);
+				return false;
 			}
-			free(file);
 		}
 	}
 
 	/* Remove the exclusive state if we're inactive */
-	if (state == RC_SERVICE_STARTED ||
-	    state == RC_SERVICE_STOPPED ||
-	    state == RC_SERVICE_INACTIVE)
-	{
-		xasprintf(&file, "%s/%s/%s", svcdir, "exclusive", base);
-		unlink(file);
-		free(file);
-	}
+	if (state == RC_SERVICE_STARTED || state == RC_SERVICE_STOPPED || state == RC_SERVICE_INACTIVE)
+		unlinkat(rc_dirfd(RC_DIR_EXCLUSIVE), base, 0);
 
 	/* Remove any options and daemons the service may have stored */
 	if (state == RC_SERVICE_STOPPED) {
-		xasprintf(&file, "%s/%s/%s", svcdir, "options", base);
-		rm_dir(file, true);
-		free(file);
-
-		xasprintf(&file, "%s/%s/%s", svcdir, "daemons", base);
-		rm_dir(file, true);
-		free(file);
-
+		rm_dir(rc_dirfd(RC_DIR_OPTIONS), base, true);
+		rm_dir(rc_dirfd(RC_DIR_DAEMONS), base, true);
 		rc_service_schedule_clear(service);
 	}
 
 	/* These are final states, so remove us from scheduled */
 	if (state == RC_SERVICE_STARTED || state == RC_SERVICE_STOPPED) {
-		xasprintf(&file, "%s/%s", svcdir, "scheduled");
-		dirs = ls_dir(file, 0);
-		TAILQ_FOREACH(dir, dirs, entries) {
-			xasprintf(&was, "%s/%s/%s", file, dir->value, base);
-			unlink(was);
-			free(was);
+		DIR *dp = do_dopendir(rc_dirfd(RC_DIR_SCHEDULED));
+		if (dp) {
+			struct dirent *d;
+			while ((d = readdir(dp))) {
+				char *was;
 
-			/* Try and remove the dir; we don't care about errors */
-			xasprintf(&was, "%s/%s", file, dir->value);
-			serrno = errno;
-			rmdir(was);
-			errno = serrno;
-			free(was);
+				xasprintf(&was, "%s/%s", d->d_name, base);
+				unlinkat(dirfd(dp), was, 0);
+				free(was);
+
+				serrno = errno;
+				unlinkat(dirfd(dp), d->d_name, AT_REMOVEDIR);
+				errno = serrno;
+			}
+
+			closedir(dp);
 		}
-		free(file);
-		rc_stringlist_free(dirs);
 	}
 	free(init);
 	return true;
@@ -913,44 +994,36 @@ rc_service_mark(const char *service, const RC_SERVICE state)
 RC_SERVICE
 rc_service_state(const char *service)
 {
-	int i;
-	int state = RC_SERVICE_STOPPED;
-	char *file;
-	RC_STRINGLIST *dirs;
-	RC_STRING *dir;
 	const char *base = basename_c(service);
-	const char *svcdir = rc_svcdir();
+	int state = RC_SERVICE_STOPPED;
+	int i;
 
 	for (i = 0; rc_service_state_names[i].name; i++) {
-		xasprintf(&file, "%s/%s/%s", svcdir, rc_service_state_names[i].name, base);
-		if (exists(file)) {
+		if (faccessat(rc_dirfd(rc_service_state_names[i].dir), base, F_OK, 0) == 0) {
 			if (rc_service_state_names[i].state <= 0x10)
 				state = rc_service_state_names[i].state;
 			else
 				state |= rc_service_state_names[i].state;
 		}
-		free(file);
 	}
 
-	if (state & RC_SERVICE_STARTED) {
+	if (state & RC_SERVICE_STARTED)
 		if (rc_service_daemons_crashed(service) && errno != EACCES)
 			state |= RC_SERVICE_CRASHED;
-	}
+
 	if (state & RC_SERVICE_STOPPED) {
-		char *sched_dir;
-		xasprintf(&sched_dir, "%s/scheduled", svcdir);
-		dirs = ls_dir(sched_dir, 0);
-		TAILQ_FOREACH(dir, dirs, entries) {
-			xasprintf(&file, "%s/scheduled/%s/%s", svcdir, dir->value, service);
-			i = exists(file);
-			free(file);
-			if (i) {
-				state |= RC_SERVICE_SCHEDULED;
-				break;
+		DIR *dp = do_dopendir(rc_dirfd(RC_DIR_SCHEDULED));
+		if (dp) {
+			struct dirent *d;
+			while ((d = readdir(dp))) {
+				if (faccessat(dirfd(dp), service, F_OK, 0) == 0) {
+					state |= RC_SERVICE_SCHEDULED;
+					break;
+				}
 			}
+
+			closedir(dp);
 		}
-		rc_stringlist_free(dirs);
-		free(sched_dir);
 	}
 
 	return state;
@@ -971,56 +1044,55 @@ rc_service_value_get(const char *service, const char *option)
 }
 
 bool
-rc_service_value_set(const char *service, const char *option,
-    const char *value)
+rc_service_value_set(const char *service, const char *option, const char *value)
 {
+	bool ret = true;
+	int optdirfd;
 	FILE *fp;
-	char *file1, *file2;
 
-	xasprintf(&file1, "%s/options/%s", rc_svcdir(), service);
-	if (mkdir(file1, 0755) != 0 && errno != EEXIST) {
-		free(file1);
+	if (mkdirat(rc_dirfd(RC_DIR_OPTIONS), service, 0755) != 0 && errno != EEXIST)
 		return false;
-	}
 
-	xasprintf(&file2, "%s/%s", file1, option);
-	free(file1);
-	if (value) {
-		fp = fopen(file2, "w");
-		free(file2);
-		if (fp == NULL)
-			return false;
+	if ((optdirfd = openat(rc_dirfd(RC_DIR_OPTIONS), service, O_RDONLY | O_DIRECTORY)) == -1)
+		return false;
+
+	if (!value) {
+		unlinkat(optdirfd, option, 0);
+	} else if ((fp = do_fopenat(optdirfd, option, O_WRONLY | O_CREAT | O_TRUNC))) {
 		fprintf(fp, "%s", value);
 		fclose(fp);
 	} else {
-		unlink(file2);
-		free(file2);
+		ret = false;
 	}
-	return true;
+
+	close(optdirfd);
+	return ret;
 }
 
 bool
 rc_service_schedule_start(const char *service, const char *service_to_start)
 {
-	char *file1, *file2;
-	char *init;
+	const char *base_to_start = basename_c(service_to_start);
+	const char *base = basename_c(service);
 	bool retval;
+	int schedfd;
+	char *init;
 
 	/* service may be a provided service, like net */
 	if (!service || !rc_service_exists(service_to_start))
 		return false;
 
-	xasprintf(&file1, "%s/scheduled/%s", rc_svcdir(), basename_c(service));
-	if (mkdir(file1, 0755) != 0 && errno != EEXIST) {
-		free(file1);
+	if (mkdirat(rc_dirfd(RC_DIR_SCHEDULED), base, 0755) != 0 && errno != EEXIST)
+		return false;
+
+	if ((schedfd = openat(rc_dirfd(RC_DIR_SCHEDULED), base, O_RDONLY)) == -1)
 		return false;
-	}
 
 	init = rc_service_resolve(service_to_start);
-	xasprintf(&file2, "%s/%s", file1, basename_c(service_to_start));
-	free(file1);
-	retval = (exists(file2) || symlink(init, file2) == 0);
-	free(file2);
+
+	retval = (faccessat(schedfd, base_to_start, F_OK, 0) == 0) || (symlinkat(init, schedfd, base_to_start) == 0);
+
+	close(schedfd);
 	free(init);
 	return retval;
 }
@@ -1028,48 +1100,32 @@ rc_service_schedule_start(const char *service, const char *service_to_start)
 bool
 rc_service_schedule_clear(const char *service)
 {
-	char *dir;
-	bool r;
-
-	xasprintf(&dir, "%s/scheduled/%s", rc_svcdir(), basename_c(service));
-	r = rm_dir(dir, true);
-	free(dir);
-	if (!r && errno == ENOENT)
-		return true;
-	return false;
+	return !rm_dir(rc_dirfd(RC_DIR_SCHEDULED), basename_c(service), true) && errno == ENOENT;
 }
 
 RC_STRINGLIST *
 rc_services_in_runlevel(const char *runlevel)
 {
-	char *dir;
 	RC_STRINGLIST *list = NULL;
 
 	if (!runlevel) {
+		const int *scriptdirfds;
 		list = rc_stringlist_new();
-		for (const char * const *dirs = rc_scriptdirs(); *dirs; dirs++) {
-			char *initd;
+		for (size_t i = 0, count = rc_scriptdirfds(&scriptdirfds); i < count; i++) {
 			RC_STRINGLIST *svcs;
 
-			xasprintf(&initd, "%s/init.d", *dirs);
-			svcs = ls_dir(initd, LS_INITD);
+			svcs = ls_dir(scriptdirfds[i], "init.d", LS_INITD);
 			TAILQ_CONCAT(list, svcs, entries);
-
 			rc_stringlist_free(svcs);
-			free(initd);
 		}
 		return list;
 	}
 
 	/* These special levels never contain any services */
-	if (strcmp(runlevel, RC_LEVEL_SINGLE) != 0) {
-		xasprintf(&dir, "%s/%s", rc_runleveldir(), runlevel);
-		list = ls_dir(dir, LS_INITD);
-		free(dir);
-	}
-	if (!list)
-		list = rc_stringlist_new();
-	return list;
+	if (strcmp(runlevel, RC_LEVEL_SINGLE) == 0)
+		return rc_stringlist_new();
+
+	return ls_dir(rc_dirfd(RC_DIR_RUNLEVEL), runlevel, LS_INITD);
 }
 
 RC_STRINGLIST *
@@ -1094,35 +1150,24 @@ rc_services_in_state(RC_SERVICE state)
 {
 	RC_STRINGLIST *services;
 	RC_STRINGLIST *list;
-	RC_STRINGLIST *dirs;
-	RC_STRING *d;
-	char *dir, *dir2;
+	struct dirent *d;
+	DIR *dp;
 
-	xasprintf(&dir, "%s/%s", rc_svcdir(), rc_parse_service_state(state));
-	if (state != RC_SERVICE_SCHEDULED) {
-		dirs = ls_dir(dir, LS_INITD);
-		free(dir);
-		return dirs;
-	}
+	if (state != RC_SERVICE_SCHEDULED)
+		return ls_dir(rc_dirfd(RC_DIR_SVCDIR), rc_parse_service_state(state), LS_INITD);
 
-	dirs = ls_dir(dir, 0);
 	list = rc_stringlist_new();
-	if (!dirs) {
-		free(dir);
+	if (!(dp = do_opendirat(rc_dirfd(RC_DIR_SVCDIR), "scheduled")))
 		return list;
-	}
 
-	TAILQ_FOREACH(d, dirs, entries) {
-		xasprintf(&dir2, "%s/%s", dir, d->value);
-		services = ls_dir(dir2, LS_INITD);
-		free(dir2);
-		if (services) {
-			TAILQ_CONCAT(list, services, entries);
-			rc_stringlist_free(services);
-		}
+	while ((d = readdir(dp))) {
+		if (!(services = ls_dir(dirfd(dp), d->d_name, LS_INITD)))
+			continue;
+		TAILQ_CONCAT(list, services, entries);
+		rc_stringlist_free(services);
 	}
-	rc_stringlist_free(dirs);
-	free(dir);
+	closedir(dp);
+
 	return list;
 }
 
@@ -1149,8 +1194,8 @@ rc_service_add(const char *runlevel, const char *service)
 	i = init = rc_service_resolve(service);
 
 	/* We need to ensure that only things in /etc/init.d are added
-	 * to the boot runlevel */
-	if (strcmp(runlevel, RC_LEVEL_BOOT) == 0) {
+	 * to the boot runlevel -- not applicable for user services. */
+	if (!rc_is_user() && strcmp(runlevel, RC_LEVEL_BOOT) == 0) {
 		path = realpath(dirname(init), NULL);
 		if (path == NULL) {
 			free(init);
@@ -1167,10 +1212,18 @@ rc_service_add(const char *runlevel, const char *service)
 		i = binit;
 	}
 
-	xasprintf(&file, "%s/%s/%s", rc_runleveldir(), runlevel,
-		basename_c(service));
+	/* We shouldn't add any temporary script to a runlevel */
+	if (strncmp(i, rc_svcdir(), strlen(rc_svcdir())) == 0) {
+		errno = ENOENT;
+		retval = false;
+		goto out;
+	}
+
+	xasprintf(&file, "%s/%s/%s", rc_runleveldir(), runlevel, basename_c(service));
 	retval = (symlink(i, file) == 0);
 	free(file);
+
+out:
 	free(binit);
 	free(init);
 	return retval;
@@ -1193,33 +1246,25 @@ RC_STRINGLIST *
 rc_services_scheduled_by(const char *service)
 {
 	RC_STRINGLIST *list = rc_stringlist_new();
-	RC_STRINGLIST *dirs;
-	RC_STRING *dir;
-	char *sched_dir;
-	char *file;
+	struct dirent *d;
+	DIR *dp;
 
-	xasprintf(&sched_dir, "%s/scheduled", rc_svcdir());
-	dirs = ls_dir(sched_dir, 0);
+	if (!(dp = do_dopendir(rc_dirfd(RC_DIR_SCHEDULED))))
+		return list;
 
-	TAILQ_FOREACH(dir, dirs, entries) {
-		xasprintf(&file, "%s/%s/%s", sched_dir, dir->value, service);
-		if (exists(file))
-			rc_stringlist_add(list, file);
+	while ((d = readdir(dp))) {
+		char *file;
+		xasprintf(&file, "%s/%s", d->d_name, service);
+		if (faccessat(dirfd(dp), file, F_OK, 0) == 0)
+			rc_stringlist_add(list, d->d_name);
 		free(file);
 	}
-	rc_stringlist_free(dirs);
-	free(sched_dir);
+
 	return list;
 }
 
 RC_STRINGLIST *
 rc_services_scheduled(const char *service)
 {
-	char *dir;
-	RC_STRINGLIST *list;
-
-	xasprintf(&dir, "%s/scheduled/%s", rc_svcdir(), basename_c(service));
-	list = ls_dir(dir, LS_INITD);
-	free(dir);
-	return list;
+	return ls_dir(rc_dirfd(RC_DIR_SCHEDULED), basename_c(service), LS_INITD);
 }
