@@ -44,8 +44,10 @@
 #include <unistd.h>
 
 #if defined(BSD) && !defined(__GNU__)
+  #if !defined(__NetBSD__)
+  #include <sys/user.h>
+  #endif
 #include <sys/param.h>
-#include <sys/user.h>
 #include <sys/sysctl.h>
 #include <kvm.h>
 #else
@@ -55,4 +57,24 @@
 #include "rc.h"
 #include "misc.h"
 
+static const char *const dirnames[RC_DIR_SYS_MAX] =
+{
+	[RC_DIR_STARTING] = "starting",
+	[RC_DIR_STARTED] = "started",
+	[RC_DIR_STOPPING] = "stopping",
+	[RC_DIR_INACTIVE] = "inactive",
+	[RC_DIR_WASINACTIVE] = "wasinactive",
+	[RC_DIR_FAILED] = "failed",
+	[RC_DIR_HOTPLUGGED] = "hotplugged",
+	[RC_DIR_DAEMONS] = "daemons",
+	[RC_DIR_OPTIONS] = "options",
+	[RC_DIR_EXCLUSIVE] = "exclusive",
+	[RC_DIR_SCHEDULED] = "scheduled",
+	[RC_DIR_INITD] = "init.d",
+	[RC_DIR_TMP] = "tmp",
+};
+
+RC_STRINGLIST *config_list(int dirfd, const char *pathname);
+void clear_dirfds(void);
+
 #endif
