@@ -35,6 +35,7 @@
 #include <time.h>
 #include <unistd.h>
 #include <stdbool.h>
+#include <inttypes.h>
 
 #if defined(__linux__) || (defined(__FreeBSD_kernel__) && defined(__GLIBC__)) \
 	|| defined(__GNU__)
@@ -48,13 +49,14 @@
 #include "einfo.h"
 #include "queue.h"
 #include "rc.h"
+#include "rc_exec.h"
 #include "misc.h"
+#include "timeutils.h"
 #include "plugin.h"
 #include "selinux.h"
 #include "_usage.h"
 #include "helpers.h"
 
-#define WAIT_INTERVAL	20000000	/* usecs to poll the lock file */
 #define WAIT_TIMEOUT	60		/* seconds until we timeout */
 #define WARN_TIMEOUT	10		/* warn about this every N seconds */
 
@@ -90,7 +92,8 @@ static RC_STRINGLIST *use_services;
 static RC_STRINGLIST *want_services;
 static RC_HOOK hook_out;
 static int exclusive_fd = -1, master_tty = -1;
-static bool sighup, skip_mark, in_background, deps, dry_run;
+static bool in_background, deps, dry_run;
+static volatile bool sighup, skip_mark, timedout;
 static pid_t service_pid;
 static int signal_pipe[2] = { -1, -1 };
 
@@ -123,8 +126,7 @@ handle_signal(int sig)
 		while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
 			if (signal_pipe[1] > -1 && pid == service_pid) {
 				if (write(signal_pipe[1], &status, sizeof(status)) == -1)
-					eerror("%s: send: %s",
-					    service, strerror(errno));
+					eerror("%s: send: %s", applet, strerror(errno));
 			}
 		}
 		break;
@@ -154,6 +156,10 @@ handle_signal(int sig)
 		exit(EXIT_FAILURE);
 		/* NOTREACHED */
 
+	case SIGALRM:
+		timedout = true;
+		break;
+
 	default:
 		eerror("%s: caught unknown signal %d", applet, sig);
 	}
@@ -165,19 +171,15 @@ handle_signal(int sig)
 static void
 unhotplug(void)
 {
-	char *file = NULL;
-
-	xasprintf(&file, "%s/hotplugged/%s", rc_svcdir(), applet);
-	if (exists(file) && unlink(file) != 0)
-		eerror("%s: unlink `%s': %s", applet, file, strerror(errno));
-	free(file);
+	if (unlinkat(rc_dirfd(RC_DIR_HOTPLUGGED), applet, 0) != 0 && errno != ENOENT)
+		eerror("%s: unlink '%s/hotplugged/%s': %s", applet, rc_svcdir(), applet, strerror(errno));
 }
 
 static void
 start_services(RC_STRINGLIST *list)
 {
 	RC_STRING *svc;
-	RC_SERVICE state = rc_service_state (service);
+	RC_SERVICE state = rc_service_state(applet);
 
 	if (!list)
 		return;
@@ -194,10 +196,8 @@ start_services(RC_STRINGLIST *list)
 			if (state & RC_SERVICE_INACTIVE ||
 			    state & RC_SERVICE_WASINACTIVE)
 			{
-				rc_service_schedule_start(service,
-				    svc->value);
-				ewarn("WARNING: %s will start when %s has started",
-				    svc->value, applet);
+				rc_service_schedule_start(applet, svc->value);
+				ewarn("WARNING: %s will start when %s has started", svc->value, applet);
 			} else
 				service_start(svc->value);
 		}
@@ -268,9 +268,9 @@ cleanup(void)
 	rc_stringlist_free(applet_list);
 	rc_stringlist_free(tmplist);
 	free(ibsave);
-	free(service);
 	free(prefix);
 	free(runlevel);
+	free(service);
 }
 
 /* Buffer and lock all output messages so that we get readable content */
@@ -286,7 +286,6 @@ write_prefix(const char *buffer, size_t bytes, bool *prefixed)
 	size_t i, j;
 	const char *ec = ecolor(ECOLOR_HILITE);
 	const char *ec_normal = ecolor(ECOLOR_NORMAL);
-	char *prefix_lock;
 	ssize_t ret = 0;
 	int fd = fileno(stdout), lock_fd = -1;
 
@@ -294,9 +293,7 @@ write_prefix(const char *buffer, size_t bytes, bool *prefixed)
 	 * Lock the prefix.
 	 * open() may fail here when running as user, as RC_SVCDIR may not be writable.
 	 */
-	xasprintf(&prefix_lock, "%s/prefix.lock", rc_svcdir());
-	lock_fd = open(prefix_lock, O_WRONLY | O_CREAT, 0664);
-	free(prefix_lock);
+	lock_fd = openat(rc_dirfd(RC_DIR_SVCDIR), "prefix.lock", O_WRONLY | O_CREAT, 0664);
 
 	if (lock_fd != -1) {
 		while (flock(lock_fd, LOCK_EX) != 0) {
@@ -340,28 +337,16 @@ write_prefix(const char *buffer, size_t bytes, bool *prefixed)
 }
 
 static int
-openrc_sh_exec(const char *openrc_sh, const char *arg1, const char *arg2)
+svc_exec(const char *command)
 {
-	if (arg2)
-		einfov("Executing: %s %s %s %s %s", openrc_sh, openrc_sh, service, arg1, arg2);
-	else
-		einfov("Executing: %s %s %s %s", openrc_sh, openrc_sh, service, arg1);
-	execl(openrc_sh, openrc_sh, service, arg1, arg2, (char *) NULL);
-	eerror("%s: exec '%s': %s", service, openrc_sh, strerror(errno));
-	_exit(EXIT_FAILURE);
-}
-
-static int
-svc_exec(const char *arg1, const char *arg2)
-{
-	int ret, fdout = fileno(stdout);
+	int ret, fdout = fileno(stdout), tmp;
 	struct termios tt;
 	struct winsize ws;
-	int i;
 	int flags = 0;
 	struct pollfd fd[2];
+	sigset_t full, old;
 	int s;
-	char *buffer;
+	char buffer[BUFSIZ];
 	size_t bytes;
 	bool prefixed = false;
 	int slave_tty;
@@ -369,12 +354,8 @@ svc_exec(const char *arg1, const char *arg2)
 	sigset_t oldmask;
 
 	/* Setup our signal pipe */
-	if (pipe(signal_pipe) == -1)
-		eerrorx("%s: pipe: %s", service, applet);
-	for (i = 0; i < 2; i++)
-		if ((flags = fcntl(signal_pipe[i], F_GETFD, 0) == -1 ||
-			fcntl(signal_pipe[i], F_SETFD, flags | FD_CLOEXEC) == -1))
-			eerrorx("%s: fcntl: %s", service, strerror(errno));
+	if (pipe2(signal_pipe, O_CLOEXEC) == -1)
+		eerrorx("%s: pipe2: %s", applet, applet);
 
 	/* Open a pty for our prefixed output
 	 * We do this instead of mapping pipes to stdout, stderr so that
@@ -398,9 +379,13 @@ svc_exec(const char *arg1, const char *arg2)
 			fcntl(slave_tty, F_SETFD, flags | FD_CLOEXEC);
 	}
 
+	sigfillset(&full);
+	/* block SIGCHLD around the fork, so service_pid is set properly in handler */
+	sigprocmask(SIG_SETMASK, &full, &old);
 	service_pid = fork();
+	sigprocmask(SIG_SETMASK, &old, NULL);
 	if (service_pid == -1)
-		eerrorx("%s: fork: %s", service, strerror(errno));
+		eerrorx("%s: fork: %s", applet, strerror(errno));
 	if (service_pid == 0) {
 		char *openrc_sh;
 		xasprintf(&openrc_sh, "%s/openrc-run.sh", rc_svcdir());
@@ -410,66 +395,63 @@ svc_exec(const char *arg1, const char *arg2)
 			dup2(slave_tty, STDERR_FILENO);
 		}
 
-		if (exists(openrc_sh))
-			openrc_sh_exec(openrc_sh, arg1, arg2);
-		else
-			openrc_sh_exec(RC_LIBEXECDIR "/sh/openrc-run.sh", arg1, arg2);
+		if (access(openrc_sh, F_OK) != 0)
+			openrc_sh = UNCONST(RC_LIBEXECDIR "/sh/openrc-run.sh");
+
+		einfov("Executing: %s %s %s", openrc_sh, service, command);
+
+		execl(openrc_sh, openrc_sh, service, command, NULL);
+		eerror("%s: exec '%s': %s", service, openrc_sh, strerror(errno));
+		_exit(EXIT_FAILURE);
 
 		/* UNREACHABLE */
 	}
 
-	buffer = xmalloc(sizeof(char) * BUFSIZ);
 	fd[0].fd = signal_pipe[0];
+	fd[1].fd = master_tty;
 	fd[0].events = fd[1].events = POLLIN;
 	fd[0].revents = fd[1].revents = 0;
-	if (master_tty >= 0) {
-		fd[1].fd = master_tty;
-		fd[1].events = POLLIN;
-		fd[1].revents = 0;
-	}
 
 	for (;;) {
-		if ((s = poll(fd, master_tty >= 0 ? 2 : 1, -1)) == -1) {
+		if (poll(fd, master_tty >= 0 ? 2 : 1, -1) == -1) {
 			if (errno != EINTR) {
-				eerror("%s: poll: %s",
-				    service, strerror(errno));
+				eerror("%s: poll: %s", applet, strerror(errno));
 				ret = -1;
 				break;
 			}
+			continue;
 		}
 
-		if (s > 0) {
-			if (fd[1].revents & (POLLIN | POLLHUP)) {
-				bytes = read(master_tty, buffer, BUFSIZ);
-				write_prefix(buffer, bytes, &prefixed);
-			}
+		if (fd[1].revents & (POLLIN | POLLHUP)) {
+			bytes = read(master_tty, buffer, BUFSIZ);
+			write_prefix(buffer, bytes, &prefixed);
+		}
 
-			/* signal_pipe receives service_pid's exit status */
-			if (fd[0].revents & (POLLIN | POLLHUP)) {
-				if ((s = read(signal_pipe[0], &ret, sizeof(ret))) != sizeof(ret)) {
-					eerror("%s: receive failed: %s", service,
-						s < 0 ? strerror(errno) : "short read");
-					ret = -1;
-					break;
-				}
-				ret = WEXITSTATUS(ret);
-				if (ret != 0 && errno == ECHILD)
-					/* killall5 -9 could cause this */
-					ret = 0;
+		/* signal_pipe receives service_pid's exit status */
+		if (fd[0].revents & (POLLIN | POLLHUP)) {
+			if ((s = read(signal_pipe[0], &ret, sizeof(ret))) != sizeof(ret)) {
+				eerror("%s: receive failed: %s", applet, s < 0 ? strerror(errno) : "short read");
+				ret = -1;
 				break;
 			}
+			ret = WEXITSTATUS(ret);
+			if (ret != 0 && errno == ECHILD)
+				/* killall5 -9 could cause this */
+				ret = 0;
+			break;
 		}
 	}
 
-	free(buffer);
-
 	sigemptyset (&sigchldmask);
 	sigaddset (&sigchldmask, SIGCHLD);
 	sigprocmask (SIG_BLOCK, &sigchldmask, &oldmask);
 
 	close(signal_pipe[0]);
-	close(signal_pipe[1]);
+	/* We should set signal_pipe[1] to -1 before closing it to avoid invalid write if
+	 * sighandler runs between close() and cleaning the variable */
+	tmp = signal_pipe[1];
 	signal_pipe[0] = signal_pipe[1] = -1;
+	close(tmp);
 
 	sigprocmask (SIG_SETMASK, &oldmask, NULL);
 
@@ -488,67 +470,47 @@ svc_exec(const char *arg1, const char *arg2)
 static bool
 svc_wait(const char *svc)
 {
-	char *file = NULL;
 	int fd;
 	bool forever = false;
 	RC_STRINGLIST *keywords;
-	struct timespec interval, timeout, warn;
+	const char *base = basename_c(svc);
+	int timeout = WAIT_TIMEOUT;
+	bool retval = false;
 
 	/* Some services don't have a timeout, like fsck */
 	keywords = rc_deptree_depend(deptree, svc, "keyword");
-	if (rc_stringlist_find(keywords, "-timeout") ||
-	    rc_stringlist_find(keywords, "notimeout"))
+	if (rc_stringlist_find(keywords, "-timeout") || rc_stringlist_find(keywords, "notimeout"))
 		forever = true;
 	rc_stringlist_free(keywords);
 
-	xasprintf(&file, "%s/exclusive/%s", rc_svcdir(), basename_c(svc));
+	if ((fd = openat(rc_dirfd(RC_DIR_EXCLUSIVE), base, O_RDONLY | O_NONBLOCK)) == -1)
+		return errno == ENOENT;
 
-	interval.tv_sec = 0;
-	interval.tv_nsec = WAIT_INTERVAL;
-	timeout.tv_sec = WAIT_TIMEOUT;
-	timeout.tv_nsec = 0;
-	warn.tv_sec = WARN_TIMEOUT;
-	warn.tv_nsec = 0;
+	timedout = false;
 	for (;;) {
-		fd = open(file, O_RDONLY | O_NONBLOCK);
-		if (fd != -1) {
-			if (flock(fd, LOCK_SH | LOCK_NB) == 0) {
-				close(fd);
-				free(file);
-				return true;
-			}
-			close(fd);
+		if (!forever)
+			alarm(WARN_TIMEOUT);
+
+		if (flock(fd, LOCK_SH) == 0) {
+			retval = true;
+			break;
 		}
-		if (errno == ENOENT) {
-			free(file);
-			return true;
-		}
-		if (errno != EWOULDBLOCK) {
-			eerror("%s: open `%s': %s", applet, file,
-			    strerror(errno));
-			free(file);
-			exit(EXIT_FAILURE);
-		}
-		if (nanosleep(&interval, NULL) == -1) {
-			if (errno != EINTR)
-				goto finish;
-		}
-		if (!forever) {
-			timespecsub(&timeout, &interval, &timeout);
-			if (timeout.tv_sec <= 0)
-				goto finish;
-			timespecsub(&warn, &interval, &warn);
-			if (warn.tv_sec <= 0) {
-				ewarn("%s: waiting for %s (%d seconds)",
-				    applet, svc, (int)timeout.tv_sec);
-				warn.tv_sec = WARN_TIMEOUT;
-				warn.tv_nsec = 0;
-			}
+
+		if (errno != EINTR)
+			break;
+
+		if (!forever && timedout) {
+			timedout = false;
+			if ((timeout -= WARN_TIMEOUT) > 0)
+				ewarn("%s: waiting for %s (%d seconds)", applet, base, timeout);
+			else
+				break;
 		}
 	}
-finish:
-	free(file);
-	return false;
+
+	alarm(0);
+	close(fd);
+	return retval;
 }
 
 static void
@@ -601,13 +563,13 @@ svc_start_check(void)
 {
 	RC_SERVICE state;
 
-	state = rc_service_state(service);
+	state = rc_service_state(applet);
 
 	if (in_background) {
 		if (!(state & (RC_SERVICE_INACTIVE | RC_SERVICE_STOPPED)))
 			exit(EXIT_FAILURE);
 		if (rc_yesno(getenv("IN_HOTPLUG")))
-			rc_service_mark(service, RC_SERVICE_HOTPLUGGED);
+			rc_service_mark(applet, RC_SERVICE_HOTPLUGGED);
 		if (strcmp(runlevel, RC_LEVEL_SYSINIT) == 0)
 			ewarnx("WARNING: %s will be started in the"
 			    " next runlevel", applet);
@@ -616,8 +578,6 @@ svc_start_check(void)
 	if (exclusive_fd == -1)
 		exclusive_fd = svc_lock(applet, !deps);
 	if (exclusive_fd == -1) {
-		if (errno == EACCES)
-			eerrorx("%s: superuser access required", applet);
 		if (state & RC_SERVICE_STOPPING)
 			ewarnx("WARNING: %s is stopping", applet);
 		else
@@ -634,7 +594,7 @@ svc_start_check(void)
 		ewarnx("WARNING: %s has already started, but is inactive",
 		    applet);
 
-	rc_service_mark(service, RC_SERVICE_STARTING);
+	rc_service_mark(applet, RC_SERVICE_STARTING);
 	hook_out = RC_HOOK_SERVICE_START_OUT;
 	rc_plugin_run(RC_HOOK_SERVICE_START_IN, applet);
 }
@@ -646,8 +606,9 @@ svc_start_deps(void)
 	RC_STRING *svc, *svc2;
 	RC_SERVICE state;
 	int depoptions = RC_DEP_TRACE;
-	size_t len;
-	char *p, *tmp;
+	size_t len, svc_count;
+	char *tmp;
+	FILE *mem;
 	pid_t pid;
 
 	errno = 0;
@@ -750,33 +711,30 @@ svc_start_deps(void)
 	if (TAILQ_FIRST(tmplist)) {
 		/* Set the state now, then unlink our exclusive so that
 		   our scheduled list is preserved */
-		rc_service_mark(service, RC_SERVICE_STOPPED);
+		rc_service_mark(applet, RC_SERVICE_STOPPED);
 
 		rc_stringlist_free(use_services);
 		use_services = NULL;
-		len = 0;
+		mem = xopen_memstream(&tmp, &len);
+		svc_count = 0;
 		TAILQ_FOREACH(svc, tmplist, entries) {
-			rc_service_schedule_start(svc->value, service);
+			rc_service_schedule_start(svc->value, applet);
 			use_services = rc_deptree_depend(deptree,
 			    "iprovide", svc->value);
 			TAILQ_FOREACH(svc2, use_services, entries)
-			    rc_service_schedule_start(svc2->value, service);
+			    rc_service_schedule_start(svc2->value, applet);
 			rc_stringlist_free(use_services);
 			use_services = NULL;
-			len += strlen(svc->value) + 2;
-		}
-
-		len += 5;
-		tmp = p = xmalloc(sizeof(char) * len);
-		TAILQ_FOREACH(svc, tmplist, entries) {
-			if (p != tmp)
-				p += snprintf(p, len, ", ");
-			p += snprintf(p, len - (p - tmp),
-			    "%s", svc->value);
+			if (svc_count > 0)
+				fputs(", ", mem);
+			fputs(svc->value, mem);
+			++svc_count;
 		}
+		xclose_memstream(mem);
 		rc_stringlist_free(tmplist);
 		tmplist = NULL;
-		ewarnx("WARNING: %s will start when %s has started", applet, tmp);
+		ewarnx("WARNING: %s will start when %s %s started",
+			applet, tmp, svc_count > 1 ? "have" : "has");
 		free(tmp);
 	}
 
@@ -796,23 +754,23 @@ static void svc_start_real(void)
 	hook_out = RC_HOOK_SERVICE_START_DONE;
 	rc_plugin_run(RC_HOOK_SERVICE_START_NOW, applet);
 	skip_mark = false;
-	started = (svc_exec("start", NULL) == 0);
+	started = (svc_exec("start") == 0);
 	if (ibsave)
 		unsetenv("IN_BACKGROUND");
 
-	if (rc_service_state(service) & RC_SERVICE_INACTIVE)
+	if (rc_service_state(applet) & RC_SERVICE_INACTIVE)
 		ewarnx("WARNING: %s has started, but is inactive", applet);
 	else if (!started)
 		eerrorx("ERROR: %s failed to start", applet);
 
 	if (!skip_mark)
-		rc_service_mark(service, RC_SERVICE_STARTED);
+		rc_service_mark(applet, RC_SERVICE_STARTED);
 	exclusive_fd = svc_unlock(applet, exclusive_fd);
 	hook_out = RC_HOOK_SERVICE_START_OUT;
 	rc_plugin_run(RC_HOOK_SERVICE_START_DONE, applet);
 
 	/* Now start any scheduled services */
-	services = rc_services_scheduled(service);
+	services = rc_services_scheduled(applet);
 	TAILQ_FOREACH(svc, services, entries)
 	    if (rc_service_state(svc->value) & RC_SERVICE_STOPPED)
 		    service_start(svc->value);
@@ -839,7 +797,7 @@ static void svc_start_real(void)
 	rc_plugin_run(RC_HOOK_SERVICE_START_OUT, applet);
 }
 
-static void
+static int
 svc_start(void)
 {
 	if (dry_run)
@@ -852,12 +810,14 @@ svc_start(void)
 		printf(" %s\n", applet);
 	else
 		svc_start_real();
+
+	return 0;
 }
 
 static int
 svc_stop_check(RC_SERVICE *state)
 {
-	*state = rc_service_state(service);
+	*state = rc_service_state(applet);
 
 	if (rc_runlevel_stopping() && *state & RC_SERVICE_FAILED)
 		exit(EXIT_FAILURE);
@@ -870,8 +830,6 @@ svc_stop_check(RC_SERVICE *state)
 	if (exclusive_fd == -1)
 		exclusive_fd = svc_lock(applet, !deps);
 	if (exclusive_fd == -1) {
-		if (errno == EACCES)
-			eerrorx("%s: superuser access required", applet);
 		if (*state & RC_SERVICE_STOPPING)
 			ewarnx("WARNING: %s is already stopping", applet);
 		eerrorx("ERROR: %s stopped by something else", applet);
@@ -884,14 +842,14 @@ svc_stop_check(RC_SERVICE *state)
 		return 1;
 	}
 
-	rc_service_mark(service, RC_SERVICE_STOPPING);
+	rc_service_mark(applet, RC_SERVICE_STOPPING);
 	hook_out = RC_HOOK_SERVICE_STOP_OUT;
 	rc_plugin_run(RC_HOOK_SERVICE_STOP_IN, applet);
 
 	if (!rc_runlevel_stopping()) {
-		if (rc_service_in_runlevel(service, RC_LEVEL_SYSINIT))
+		if (rc_service_in_runlevel(applet, RC_LEVEL_SYSINIT))
 			ewarn("WARNING: you are stopping a sysinit service");
-		else if (rc_service_in_runlevel(service, RC_LEVEL_BOOT))
+		else if (rc_service_in_runlevel(applet, RC_LEVEL_BOOT))
 			ewarn("WARNING: you are stopping a boot service");
 	}
 
@@ -961,14 +919,14 @@ svc_stop_deps(RC_SERVICE state)
 			continue;
 		if (rc_runlevel_stopping()) {
 			/* If shutting down, we should stop even
-			 * if a dependant failed */
+			 * if a dependent failed */
 			if (runlevel &&
 			    (strcmp(runlevel,
 				RC_LEVEL_SHUTDOWN) == 0 ||
 				strcmp(runlevel,
 				    RC_LEVEL_SINGLE) == 0))
 				continue;
-			rc_service_mark(service, RC_SERVICE_FAILED);
+			rc_service_mark(applet, RC_SERVICE_FAILED);
 		}
 		eerrorx("ERROR: cannot stop %s as %s "
 		    "is still up", applet, svc->value);
@@ -1004,7 +962,7 @@ svc_stop_real(void)
 	hook_out = RC_HOOK_SERVICE_STOP_DONE;
 	rc_plugin_run(RC_HOOK_SERVICE_STOP_NOW, applet);
 	skip_mark = false;
-	stopped = (svc_exec("stop", NULL) == 0);
+	stopped = (svc_exec("stop") == 0);
 	if (ibsave)
 		unsetenv("IN_BACKGROUND");
 
@@ -1013,9 +971,9 @@ svc_stop_real(void)
 
 	if (!skip_mark) {
 		if (in_background)
-		    rc_service_mark(service, RC_SERVICE_INACTIVE);
+		    rc_service_mark(applet, RC_SERVICE_INACTIVE);
 		else
-		    rc_service_mark(service, RC_SERVICE_STOPPED);
+		    rc_service_mark(applet, RC_SERVICE_STOPPED);
 	}
 	hook_out = RC_HOOK_SERVICE_STOP_OUT;
 	rc_plugin_run(RC_HOOK_SERVICE_STOP_DONE, applet);
@@ -1044,10 +1002,10 @@ svc_stop(void)
 	return 0;
 }
 
-static void
+static int
 svc_restart(void)
 {
-	if (!(rc_service_state(service) & RC_SERVICE_STOPPED)) {
+	if (!(rc_service_state(applet) & RC_SERVICE_STOPPED)) {
 		get_started_services();
 		svc_stop();
 		if (dry_run)
@@ -1059,6 +1017,8 @@ svc_restart(void)
 	start_services(restart_services);
 	rc_stringlist_free(restart_services);
 	restart_services = NULL;
+
+	return 0;
 }
 
 static bool
@@ -1091,21 +1051,137 @@ service_plugable(void)
 	return allow;
 }
 
+static int
+simple_command(void)
+{
+	char *save = prefix;
+	int retval;
+
+	eprefix(NULL);
+	prefix = NULL;
+
+	retval = svc_exec(optarg);
+
+	eprefix(save);
+	prefix = save;
+
+	return retval;
+}
+
+static int
+resolve_deptype(void)
+{
+	int depoptions = RC_DEP_TRACE;
+	RC_STRING *svc;
+
+	errno = 0;
+	if (rc_conf_yesno("rc_depend_strict") || errno == ENOENT)
+		depoptions |= RC_DEP_STRICT;
+
+	if (!deptree && ((deptree = _rc_deptree_load(0, NULL)) == NULL))
+		eerrorx("failed to load deptree");
+
+	tmplist = rc_stringlist_new();
+	rc_stringlist_add(tmplist, optarg);
+	services = rc_deptree_depends(deptree, tmplist, applet_list, runlevel, depoptions);
+	rc_stringlist_free(tmplist);
+
+	tmplist = NULL;
+	TAILQ_FOREACH(svc, services, entries)
+		printf("%s ", svc->value);
+	printf ("\n");
+
+	rc_stringlist_free(services);
+	services = NULL;
+
+	return 0;
+}
+
+static int
+cmd_condrestart(void)
+{
+	if (rc_service_state(applet) & RC_SERVICE_STARTED)
+		svc_restart();
+	return 0;
+}
+
+static int
+cmd_stop(void)
+{
+	RC_STRING *svc;
+
+	if (strcmp(optarg, "pause") == 0) {
+		ewarn("WARNING: 'pause' is deprecated; please use '--nodeps stop'");
+		deps = false;
+	}
+
+	if (deps && in_background)
+		get_started_services();
+
+	if (svc_stop() == 1)
+		return 0; /* Service has been stopped already */
+
+	if (deps) {
+		if (!in_background && !rc_runlevel_stopping() && rc_service_state(applet) & RC_SERVICE_STOPPED)
+			unhotplug();
+
+		if (in_background && rc_service_state(applet) & RC_SERVICE_INACTIVE) {
+			TAILQ_FOREACH(svc, restart_services, entries)
+				if (rc_service_state(svc->value) & RC_SERVICE_STOPPED)
+					rc_service_schedule_start(applet, svc->value);
+		}
+	}
+
+	return 0;
+}
+
+static int
+cmd_zap(void)
+{
+	einfo("Manually resetting %s to stopped state", applet);
+	if (!rc_service_mark(applet, RC_SERVICE_STOPPED))
+		eerrorx("rc_service_mark: %s", strerror(errno));
+	unhotplug();
+
+	return 0;
+}
+
+struct {
+	const char *name;
+	int (*handler)(void);
+	bool reset_restart_list;
+} commands[] = {
+	{ "describe",           simple_command,  false },
+	{ "help",               simple_command,  false },
+	{ "depend",             simple_command,  false },
+	{ "status",             simple_command,  false },
+	{ "ineed",              resolve_deptype, false },
+	{ "iuse",               resolve_deptype, false },
+	{ "iwant",              resolve_deptype, false },
+	{ "needsme",            resolve_deptype, false },
+	{ "usesme",             resolve_deptype, false },
+	{ "wantsme",            resolve_deptype, false },
+	{ "iafter",             resolve_deptype, false },
+	{ "ibefore",            resolve_deptype, false },
+	{ "iprovide",           resolve_deptype, false },
+	{ "start",              svc_start,       true  },
+	{ "stop",               cmd_stop,        true  },
+	{ "restart",            svc_restart,     true  },
+	{ "condrestart",        cmd_condrestart, true  },
+	{ "conditionalrestart", cmd_condrestart, true  },
+	{ "zap",                cmd_zap,         true  },
+};
+
 int main(int argc, char **argv)
 {
-	bool doneone = false;
-	int retval, opt, depoptions = RC_DEP_TRACE;
-	RC_STRING *svc;
-	char *path = NULL;
-	char *lnk = NULL;
-	char *dir, *save = NULL, *saveLnk = NULL;
+	int retval, opt;
+	const char *workingdir = "/";
 	char *pidstr = NULL;
 	size_t l = 0, ll;
-	const char *file;
 	struct stat stbuf;
 
 	/* Show help if insufficient args */
-	if (argc < 2 || !exists(argv[1])) {
+	if (argc < 2 || access(argv[1], F_OK) != 0) {
 		fprintf(stderr, "openrc-run should not be run directly\n");
 		exit(EXIT_FAILURE);
 	}
@@ -1113,100 +1189,20 @@ int main(int argc, char **argv)
 	applet = basename_c(argv[0]);
 
 	if (stat(argv[1], &stbuf) != 0) {
-		fprintf(stderr, "openrc-run `%s': %s\n",
-		    argv[1], strerror(errno));
+		fprintf(stderr, "openrc-run `%s': %s\n", argv[1], strerror(errno));
 		exit(EXIT_FAILURE);
 	}
 
-	atexit(cleanup);
+	if (rc_yesno(getenv("RC_USER_SERVICES")))
+		rc_set_user();
 
-	/* We need to work out the real full path to our service.
-	 * This works fine, provided that we ONLY allow multiplexed services
-	 * to exist in the same directory as the master link.
-	 * Also, the master link as to be a real file in the init dir. */
-	path = realpath(argv[1], NULL);
-	if (!path) {
-		fprintf(stderr, "realpath: %s\n", strerror(errno));
-		exit(EXIT_FAILURE);
-	}
-	lnk = xmalloc(4096);
-	memset(lnk, 0, 4096);
-	if (readlink(argv[1], lnk, 4096-1)) {
-		dir = dirname(path);
-		if (strchr(lnk, '/')) {
-			save = xstrdup(dir);
-			saveLnk = xstrdup(lnk);
-			dir = dirname(saveLnk);
-			if (strcmp(dir, save) == 0)
-				file = basename_c(argv[1]);
-			else
-				file = basename_c(lnk);
-			dir = save;
-		} else
-			file = basename_c(argv[1]);
-		xasprintf(&service, "%s/%s", dir, file);
-		if (stat(service, &stbuf) != 0) {
-			free(service);
-			service = xstrdup(lnk);
-		}
-		free(save);
-		free(saveLnk);
-	}
-	free(lnk);
-	if (!service)
-		service = xstrdup(path);
-	applet = basename_c(service);
+	if (!(service = (realpath(argv[1], NULL))))
+		eerrorx("readpath: %s", strerror(errno));
+	applet = basename_c(argv[1]);
 
 	if (argc < 3)
 		usage(EXIT_FAILURE);
 
-	/* Change dir to / to ensure all init scripts don't use stuff in pwd */
-	if (chdir("/") == -1)
-		eerror("chdir: %s", strerror(errno));
-
-	if ((runlevel = xstrdup(getenv("RC_RUNLEVEL"))) == NULL) {
-		env_filter();
-		env_config();
-		runlevel = rc_runlevel_get();
-	}
-
-	setenv("EINFO_LOG", service, 1);
-	setenv("RC_SVCNAME", applet, 1);
-
-	/* Set an env var so that we always know our pid regardless of any
-	   subshells the init script may create so that our mark_service_*
-	   functions can always instruct us of this change */
-	xasprintf(&pidstr, "%d", (int) getpid());
-	setenv("RC_OPENRC_PID", pidstr, 1);
-	/*
-	 * RC_RUNSCRIPT_PID is deprecated, but we will keep it for a while
-	 * for safety.
-	 */
-	setenv("RC_RUNSCRIPT_PID", pidstr, 1);
-
-	/* eprefix is kinda klunky, but it works for our purposes */
-	if (rc_conf_yesno("rc_parallel")) {
-		/* Get the longest service name */
-		services = rc_services_in_runlevel(NULL);
-		TAILQ_FOREACH(svc, services, entries) {
-			ll = strlen(svc->value);
-			if (ll > l)
-				l = ll;
-		}
-		rc_stringlist_free(services);
-		services = NULL;
-		ll = strlen(applet);
-		if (ll > l)
-			l = ll;
-
-		/* Make our prefix string */
-		prefix = xmalloc(sizeof(char) * l + 1);
-		memcpy(prefix, applet, ll);
-		memset(prefix + ll, ' ', l - ll);
-		memset(prefix + l, 0, 1);
-		eprefix(prefix);
-	}
-
 	/* Ok, we are ready to go, so setup selinux if applicable */
 	selinux_setup(argv);
 
@@ -1229,11 +1225,11 @@ int main(int argc, char **argv)
 			    fcntl(exclusive_fd, F_GETFD, 0) | FD_CLOEXEC);
 			break;
 		case 's':
-			if (!(rc_service_state(service) & RC_SERVICE_STARTED))
+			if (!(rc_service_state(applet) & RC_SERVICE_STARTED))
 				exit(EXIT_FAILURE);
 			break;
 		case 'S':
-			if (!(rc_service_state(service) & RC_SERVICE_STOPPED))
+			if (!(rc_service_state(applet) & RC_SERVICE_STOPPED))
 				exit(EXIT_FAILURE);
 			break;
 		case 'D':
@@ -1245,6 +1241,60 @@ int main(int argc, char **argv)
 		case_RC_COMMON_GETOPT
 		}
 
+	/* Change dir to / to ensure all init scripts don't use stuff in pwd
+	 * For user services, change to the user's HOME instead. */
+	if (rc_is_user() && !(workingdir = getenv("HOME")))
+		eerrorx("HOME is unset.");
+
+	if (chdir(workingdir) == -1)
+		eerror("chdir: %s", strerror(errno));
+
+	atexit(cleanup);
+
+	if ((runlevel = xstrdup(getenv("RC_RUNLEVEL"))) == NULL) {
+		env_filter();
+		env_config();
+		runlevel = rc_runlevel_get();
+	}
+
+	setenv("EINFO_LOG", applet, 1);
+	setenv("RC_SVCNAME", applet, 1);
+
+	/* Set an env var so that we always know our pid regardless of any
+	   subshells the init script may create so that our mark_service_*
+	   functions can always instruct us of this change */
+	xasprintf(&pidstr, "%d", (int) getpid());
+	setenv("RC_OPENRC_PID", pidstr, 1);
+	/*
+	 * RC_RUNSCRIPT_PID is deprecated, but we will keep it for a while
+	 * for safety.
+	 */
+	setenv("RC_RUNSCRIPT_PID", pidstr, 1);
+
+	/* eprefix is kinda klunky, but it works for our purposes */
+	if (rc_conf_yesno("rc_parallel")) {
+		RC_STRING *svc;
+		/* Get the longest service name */
+		services = rc_services_in_runlevel(NULL);
+		TAILQ_FOREACH(svc, services, entries) {
+			ll = strlen(svc->value);
+			if (ll > l)
+				l = ll;
+		}
+		rc_stringlist_free(services);
+		services = NULL;
+		ll = strlen(applet);
+		if (ll > l)
+			l = ll;
+
+		/* Make our prefix string */
+		prefix = xmalloc(sizeof(char) * l + 1);
+		memcpy(prefix, applet, ll);
+		memset(prefix + ll, ' ', l - ll);
+		memset(prefix + l, 0, 1);
+		eprefix(prefix);
+	}
+
 	if (rc_yesno(getenv("RC_NODEPS")))
 		deps = false;
 
@@ -1272,6 +1322,7 @@ int main(int argc, char **argv)
 
 	/* Setup a signal handler */
 	signal_setup(SIGHUP, handle_signal);
+	signal_setup(SIGALRM, handle_signal);
 	signal_setup(SIGUSR1, handle_signal);
 	signal_setup(SIGINT, handle_signal);
 	signal_setup(SIGQUIT, handle_signal);
@@ -1284,9 +1335,15 @@ int main(int argc, char **argv)
 	applet_list = rc_stringlist_new();
 	rc_stringlist_add(applet_list, applet);
 
+	if (optind >= argc)
+		usage(EXIT_FAILURE);
+
 	/* Now run each option */
 	retval = EXIT_SUCCESS;
+
+next_cmd:
 	while (optind < argc) {
+		int cmdval;
 		optarg = argv[optind++];
 
 		/* Abort on a sighup here */
@@ -1300,112 +1357,27 @@ int main(int argc, char **argv)
 		unsetenv("RC_CMD");
 		setenv("RC_CMD", optarg, 1);
 
-		doneone = true;
+		for (size_t i = 0; i < ARRAY_SIZE(commands); i++) {
+			if (strcmp(optarg, commands[i].name) != 0)
+				continue;
 
-		if (strcmp(optarg, "describe") == 0 ||
-		    strcmp(optarg, "help") == 0 ||
-		    strcmp(optarg, "depend") == 0)
-		{
-			save = prefix;
-			eprefix(NULL);
-			prefix = NULL;
-			svc_exec(optarg, NULL);
-			eprefix(save);
-			prefix = save;
-		} else if (strcmp(optarg, "ineed") == 0 ||
-		    strcmp(optarg, "iuse") == 0 ||
-		    strcmp(optarg, "iwant") == 0 ||
-		    strcmp(optarg, "needsme") == 0 ||
-		    strcmp(optarg, "usesme") == 0 ||
-		    strcmp(optarg, "wantsme") == 0 ||
-		    strcmp(optarg, "iafter") == 0 ||
-		    strcmp(optarg, "ibefore") == 0 ||
-		    strcmp(optarg, "iprovide") == 0)
-		{
-			errno = 0;
-			if (rc_conf_yesno("rc_depend_strict") ||
-			    errno == ENOENT)
-				depoptions |= RC_DEP_STRICT;
+			if ((cmdval = commands[i].handler()) != 0)
+				retval = cmdval;
 
-			if (!deptree &&
-			    ((deptree = _rc_deptree_load(0, NULL)) == NULL))
-				eerrorx("failed to load deptree");
+			if (commands[i].reset_restart_list) {
+				/* We should ensure this list is empty after
+				 * an action is done */
+				rc_stringlist_free(restart_services);
+				restart_services = NULL;
+			}
 
-			tmplist = rc_stringlist_new();
-			rc_stringlist_add(tmplist, optarg);
-			services = rc_deptree_depends(deptree, tmplist,
-			    applet_list,
-			    runlevel, depoptions);
-			rc_stringlist_free(tmplist);
-			tmplist = NULL;
-			TAILQ_FOREACH(svc, services, entries)
-			    printf("%s ", svc->value);
-			printf ("\n");
-			rc_stringlist_free(services);
-			services = NULL;
-		} else if (strcmp (optarg, "status") == 0) {
-			save = prefix;
-			eprefix(NULL);
-			prefix = NULL;
-			retval = svc_exec("status", NULL);
-		} else {
-			if (strcmp(optarg, "conditionalrestart") == 0 ||
-			    strcmp(optarg, "condrestart") == 0)
-			{
-				if (rc_service_state(service) &
-				    RC_SERVICE_STARTED)
-					svc_restart();
-			} else if (strcmp(optarg, "restart") == 0) {
-				svc_restart();
-			} else if (strcmp(optarg, "start") == 0) {
-				svc_start();
-			} else if (strcmp(optarg, "stop") == 0 || strcmp(optarg, "pause") == 0) {
-				if (strcmp(optarg, "pause") == 0) {
-					ewarn("WARNING: 'pause' is deprecated; please use '--nodeps stop'");
-					deps = false;
-				}
-				if (deps && in_background)
-					get_started_services();
-				if (svc_stop() == 1)
-					continue; /* Service has been stopped already */
-				if (deps) {
-					if (!in_background &&
-					    !rc_runlevel_stopping() &&
-					    rc_service_state(service) &
-					    RC_SERVICE_STOPPED)
-						unhotplug();
-
-					if (in_background &&
-					    rc_service_state(service) &
-					    RC_SERVICE_INACTIVE)
-					{
-						TAILQ_FOREACH(svc,
-						    restart_services,
-						    entries)
-						    if (rc_service_state(svc->value) &
-							RC_SERVICE_STOPPED)
-							    rc_service_schedule_start(service, svc->value);
-					}
-				}
-			} else if (strcmp(optarg, "zap") == 0) {
-				einfo("Manually resetting %s to stopped state",
-				    applet);
-				if (!rc_service_mark(applet,
-					RC_SERVICE_STOPPED))
-					eerrorx("rc_service_mark: %s",
-					    strerror(errno));
-				unhotplug();
-			} else
-				retval = svc_exec(optarg, NULL);
-
-			/* We should ensure this list is empty after
-			 * an action is done */
-			rc_stringlist_free(restart_services);
-			restart_services = NULL;
+			goto next_cmd;
 		}
 
-		if (!doneone)
-			usage(EXIT_FAILURE);
+		if ((cmdval = svc_exec(optarg)) != 0)
+			retval = cmdval;
+		rc_stringlist_free(restart_services);
+		restart_services = NULL;
 	}
 
 	return retval;
