@@ -71,15 +71,15 @@ const char * const longopts_help[] = {
 	"write wtmp boot record and exit",
 	longopts_help_COMMON
 };
-const char *usagestring = "" \
-						   "Usage: openrc-shutdown -c | --cancel\n" \
-						   "   or: openrc-shutdown -R | --reexec\n" \
-						   "   or: openrc-shutdown -w | --write-only\n" \
-						   "   or: openrc-shutdown -H | --halt time\n" \
-						   "   or: openrc-shutdown -K | --kexec time\n" \
-						   "   or: openrc-shutdown -p | --poweroff time\n" \
-						   "   or: openrc-shutdown -r | --reboot time\n" \
-						   "   or: openrc-shutdown -s | --single time";
+const char *usagestring = ""
+	"Usage: openrc-shutdown -c | --cancel\n"
+	"   or: openrc-shutdown -R | --reexec\n"
+	"   or: openrc-shutdown -w | --write-only\n"
+	"   or: openrc-shutdown -H | --halt time\n"
+	"   or: openrc-shutdown -K | --kexec time\n"
+	"   or: openrc-shutdown -p | --poweroff time\n"
+	"   or: openrc-shutdown -r | --reboot time\n"
+	"   or: openrc-shutdown -s | --single time";
 const char *exclusive = "Select one of "
 	"--cancel, --halt, --kexec, --poweroff, --reexec, --reboot, --single or \n"
 	"--write-only";
@@ -333,7 +333,7 @@ int main(int argc, char **argv)
 	unlink(nologin_file);
 	unlink(shutdown_pid);
 	if (do_halt) {
-		if (exists("/run/initctl")) {
+		if (access("/run/initctl", F_OK) == 0) {
 			sysvinit_setenv("INIT_HALT", "HALT");
 			sysvinit_runlevel('0');
 		} else
@@ -341,18 +341,18 @@ int main(int argc, char **argv)
 	} else if (do_kexec)
 		send_cmd("kexec");
 	else if (do_poweroff) {
-		if (exists("/run/initctl")) {
+		if (access("/run/initctl", F_OK) == 0) {
 			sysvinit_setenv("INIT_HALT", "POWEROFF");
 			sysvinit_runlevel('0');
 		} else
 			send_cmd("poweroff");
 	} else if (do_reboot) {
-		if (exists("/run/initctl"))
+		if (access("/run/initctl", F_OK) == 0)
 			sysvinit_runlevel('6');
 		else
 			send_cmd("reboot");
 	} else if (do_single) {
-		if (exists("/run/initctl"))
+		if (access("/run/initctl", F_OK) == 0)
 			sysvinit_runlevel('S');
 		else
 			send_cmd("single");
