@@ -110,6 +110,20 @@ write_time(FILE *f, const char *s)
 	fflush(f);
 }
 
+static void
+mkpath(const char *path)
+{
+	char *dir = xstrdup(path), *slash = dir;
+	while ((slash = strchr(slash + 1, '/'))) {
+		*slash = '\0';
+		if (mkdir(path, 0755) == -1 && errno != EEXIST)
+			break;
+		*slash = '/';
+	}
+	mkdir(path, 0755);
+	free(dir);
+}
+
 void
 rc_logger_close(void)
 {
@@ -141,22 +155,17 @@ rc_logger_open(const char *level)
 	struct pollfd fd[2];
 	int s = 0;
 	size_t bytes;
-	int i;
 	FILE *log = NULL;
 	FILE *plog = NULL;
 	const char *logfile;
-	char *tmplog;
+	char *tmplog, *usrlog = NULL;
 	int log_error = 0;
 
 	if (!rc_conf_yesno("rc_logger"))
 		return;
 
-	if (pipe(signal_pipe) == -1)
-		eerrorx("pipe: %s", strerror(errno));
-	for (i = 0; i < 2; i++)
-		if ((s = fcntl (signal_pipe[i], F_GETFD, 0) == -1 ||
-			fcntl (signal_pipe[i], F_SETFD, s | FD_CLOEXEC) == -1))
-			eerrorx("fcntl: %s", strerror (errno));
+	if (pipe2(signal_pipe, O_CLOEXEC) == -1)
+		eerrorx("pipe2: %s", strerror(errno));
 
 	if (isatty(STDOUT_FILENO)) {
 		tcgetattr(STDOUT_FILENO, &tt);
@@ -248,8 +257,21 @@ rc_logger_open(const char *level)
 
 		/* Append the temporary log to the real log */
 		logfile = rc_conf_value("rc_log_path");
-		if (logfile == NULL)
-			logfile = DEFAULTLOG;
+		if (logfile == NULL) {
+			if (!rc_is_user()) {
+				logfile = DEFAULTLOG;
+			} else {
+				const char *state_home = getenv("XDG_STATE_HOME");
+				if (state_home) {
+					xasprintf(&usrlog, "%s/rc.log", state_home);
+				} else if ((state_home = getenv("HOME"))) {
+					xasprintf(&usrlog, "%s/.local/state/rc.log", state_home);
+					state_home = ".local/state";
+				}
+				mkpath(state_home);
+				logfile = usrlog;
+			}
+		}
 		if (!strcmp(logfile, tmplog)) {
 			eerror("Cowardly refusing to concatenate a logfile into itself.");
 			eerrorx("Please change rc_log_path to something other than %s to get rid of this message", tmplog);
@@ -282,12 +304,14 @@ rc_logger_open(const char *level)
 			}
 		}
 
+		free(usrlog);
+
 		/* Try to keep the temporary log in case of errors */
 		if (!log_error) {
 			if (errno != EROFS && ((strcmp(level, RC_LEVEL_SHUTDOWN) != 0) && (strcmp(level, RC_LEVEL_SYSINIT) != 0)))
 				if (unlink(tmplog) == -1)
 					eerror("Error: unlink(%s) failed: %s", tmplog, strerror(errno));
-		} else if (exists(tmplog)) {
+		} else if (access(tmplog, F_OK) == 0) {
 			eerrorx("Warning: temporary logfile left behind: %s", tmplog);
 		}
 
