@@ -13,6 +13,9 @@
 #ifndef RC_LOGGER_H
 #define RC_LOGGER_H
 
+#include <stdbool.h>
+#include <sys/types.h>
+
 extern pid_t rc_logger_pid;
 extern int rc_logger_tty;
 extern bool rc_in_logger;
