@@ -41,6 +41,7 @@
 #include "einfo.h"
 #include "queue.h"
 #include "rc.h"
+#include "rc_exec.h"
 #include "rc-logger.h"
 #include "misc.h"
 #include "plugin.h"
@@ -64,8 +65,8 @@ const char * const longopts_help[] = {
 	"output the RC system type, if any",
 	longopts_help_COMMON
 };
-const char *usagestring = ""					\
-    "Usage: openrc [options] [<runlevel>]";
+const char *usagestring = ""
+	"Usage: openrc [options] [<runlevel>]";
 
 #define INITSH                  RC_LIBEXECDIR "/sh/init.sh"
 #define INITEARLYSH             RC_LIBEXECDIR "/sh/init-early.sh"
@@ -92,37 +93,26 @@ clean_failed(void)
 {
 	DIR *dp;
 	struct dirent *d;
-	char *path;
-	char *failed_dir;
-
-	xasprintf(&failed_dir, "%s/failed", rc_svcdir());
 
 	/* Clean the failed services state dir now */
-	if ((dp = opendir(failed_dir))) {
-		while ((d = readdir(dp))) {
-			if (d->d_name[0] == '.' &&
-			    (d->d_name[1] == '\0' ||
-				(d->d_name[1] == '.' && d->d_name[2] == '\0')))
-				continue;
+	if (!(dp = do_opendirat(rc_dirfd(RC_DIR_SVCDIR), "failed")))
+		return;
 
-			xasprintf(&path, "%s/%s", failed_dir, d->d_name);
-			if (unlink(path))
-				eerror("%s: unlink `%s': %s",
-				    applet, path, strerror(errno));
-			free(path);
-		}
-		closedir(dp);
+	while ((d = readdir(dp))) {
+		if (strcmp(d->d_name, ".") == 0 || strcmp(d->d_name, "..") == 0)
+			continue;
+
+		if (unlinkat(dirfd(dp), d->d_name, 0) == -1)
+			eerror("%s: unlink '%s/failed/%s': %s", applet, rc_svcdir(), d->d_name, strerror(errno));
 	}
-
-	free(failed_dir);
+	closedir(dp);
 }
 
 static void
 cleanup(void)
 {
+	int svcdirfd = rc_dirfd(RC_DIR_SVCDIR);
 	RC_PID *p, *tmp;
-	const char *svcdir = rc_svcdir();
-	static const char *subdirs[] = { "rc.starting", "rc.stopping" };
 
 	if (!rc_in_logger && !rc_in_plugin &&
 	    applet && (strcmp(applet, "rc") == 0 || strcmp(applet, "openrc") == 0))
@@ -138,12 +128,8 @@ cleanup(void)
 		}
 
 		/* Clean runlevel start, stop markers */
-		for (size_t i = 0; i < ARRAY_SIZE(subdirs); i++) {
-			char *path;
-			xasprintf(&path, "%s/%s", svcdir, subdirs[i]);
-			rmdir(path);
-			free(path);
-		}
+		unlinkat(svcdirfd, "rc.starting", AT_REMOVEDIR);
+		unlinkat(svcdirfd, "rc.stopping", AT_REMOVEDIR);
 
 		clean_failed();
 		rc_logger_close();
@@ -222,13 +208,9 @@ static void
 mark_interactive(void)
 {
 	FILE *fp;
-	char *dir;
 
-	xasprintf(&dir, "%s/interactive", rc_svcdir());
-	if ((fp = fopen(dir, "w")))
+	if ((fp = do_fopenat(rc_dirfd(RC_DIR_SVCDIR), "interactive", O_WRONLY | O_CREAT | O_TRUNC)))
 		fclose(fp);
-
-	free(dir);
 }
 
 static void
@@ -315,15 +297,16 @@ set_krunlevel(const char *level)
 {
 	FILE *fp;
 
+	if (rc_is_user())
+		return true;
+
 	if (!level ||
 	    strcmp(level, getenv ("RC_BOOTLEVEL")) == 0 ||
 	    strcmp(level, RC_LEVEL_SINGLE) == 0 ||
 	    strcmp(level, RC_LEVEL_SYSINIT) == 0)
 	{
-		if (exists(RC_KRUNLEVEL) &&
-		    unlink(RC_KRUNLEVEL) != 0)
-			eerror("unlink `%s': %s", RC_KRUNLEVEL,
-			    strerror(errno));
+		if (access(RC_KRUNLEVEL, F_OK) == 0 && unlink(RC_KRUNLEVEL) != 0)
+			eerror("unlink `%s': %s", RC_KRUNLEVEL, strerror(errno));
 		return false;
 	}
 
@@ -343,7 +326,10 @@ static char *get_krunlevel(void)
 	FILE *fp;
 	size_t i = 0;
 
-	if (!exists(RC_KRUNLEVEL))
+	if (rc_is_user())
+		return NULL;
+
+	if (access(RC_KRUNLEVEL, F_OK) != 0)
 		return NULL;
 	if (!(fp = fopen(RC_KRUNLEVEL, "r"))) {
 		eerror("fopen `%s': %s", RC_KRUNLEVEL, strerror(errno));
@@ -497,21 +483,23 @@ do_sysinit(void)
 	/* exec init-early.sh if it exists
 	 * This should just setup the console to use the correct
 	 * font. Maybe it should setup the keyboard too? */
-	if (exists(INITEARLYSH))
+	if (access(INITEARLYSH, F_OK) == 0)
 		run_program(INITEARLYSH);
 
 	uname(&uts);
-	printf("\n   %sOpenRC %s" VERSION "%s is starting up %s",
-	    ecolor(ECOLOR_GOOD), ecolor(ECOLOR_HILITE),
-	    ecolor(ECOLOR_NORMAL), ecolor(ECOLOR_BRACKET));
+	if (!rc_yesno(getenv("EINFO_QUIET"))) {
+		printf("\n   %sOpenRC %s" VERSION "%s is starting up %s",
+		    ecolor(ECOLOR_GOOD), ecolor(ECOLOR_HILITE),
+		    ecolor(ECOLOR_NORMAL), ecolor(ECOLOR_BRACKET));
 #ifdef BRANDING
-	printf(BRANDING " (%s)", uts.machine);
+		printf(BRANDING " (%s)", uts.machine);
 #else
-	printf("%s %s (%s)",
-	    uts.sysname,
-	    uts.release,
-	    uts.machine);
+		printf("%s %s (%s)",
+		    uts.sysname,
+		    uts.release,
+		    uts.machine);
 #endif
+	}
 
 	if ((sys = rc_sys()))
 		printf(" [%s]", sys);
@@ -530,8 +518,7 @@ do_sysinit(void)
 	 * sys */
 	if ((sys = rc_sys()))
 		setenv("RC_SYS", sys, 1);
-	/* force an update of the dependency tree */
-	if ((main_deptree = _rc_deptree_load(1, NULL)) == NULL)
+	if ((main_deptree = _rc_deptree_load(false, NULL)) == NULL)
 		eerrorx("failed to load deptree");
 }
 
@@ -544,7 +531,7 @@ runlevel_config(const char *service, const char *level)
 
 	dir = dirname(dirname(init));
 	xasprintf(&conf, "%s/conf.d/%s.%s", dir, service, level);
-	retval = exists(conf);
+	retval = access(conf, F_OK) == 0;
 	free(conf);
 	free(init);
 	return retval;
@@ -668,11 +655,9 @@ do_start_services(const RC_STRINGLIST *start_services, bool parallel)
 	bool interactive = false;
 	RC_SERVICE state;
 	bool crashed = false;
-	char *interactive_dir;
 
-	xasprintf(&interactive_dir, "%s/interactive", rc_svcdir());
 	if (!rc_yesno(getenv("EINFO_QUIET")))
-		interactive = exists(interactive_dir);
+		interactive = faccessat(rc_dirfd(RC_DIR_SVCDIR), "interactive", F_OK, 0) == 0;
 	errno = 0;
 	crashed = rc_conf_yesno("rc_crashed_start");
 	if (errno == ENOENT)
@@ -731,9 +716,7 @@ do_start_services(const RC_STRINGLIST *start_services, bool parallel)
 			 strcmp(runlevel, getenv("RC_BOOTLEVEL")) == 0))
 		mark_interactive();
 	else
-		if (exists(interactive_dir))
-			unlink(interactive_dir);
-	free(interactive_dir);
+		unlinkat(rc_dirfd(RC_DIR_SVCDIR), "interactive", 0);
 }
 
 #ifdef RC_DEBUG
@@ -772,10 +755,8 @@ int main(int argc, char **argv)
 	RC_STRING *service;
 	bool going_down = false;
 	int depoptions = RC_DEP_STRICT | RC_DEP_TRACE;
-	const char *svcdir = rc_svcdir();
-	char *rc_starting, *rc_stopping;
-	char *deptree_skewed;
 	char *krunlevel = NULL;
+	const char *workingdir = "/";
 	char *pidstr = NULL;
 	int opt;
 	bool parallel;
@@ -796,24 +777,14 @@ int main(int argc, char **argv)
 	applet = basename_c(argv[0]);
 	LIST_INIT(&service_pids);
 	LIST_INIT(&free_these_pids);
-	atexit(cleanup);
 	if (!applet)
 		eerrorx("arguments required");
 
 	argc--;
 	argv++;
 
-	/* Change dir to / to ensure all scripts don't use stuff in pwd */
-	if (chdir("/") == -1)
-		eerror("chdir: %s", strerror(errno));
-
-	/* Ensure our environment is pure
-	 * Also, add our configuration to it */
-	env_filter();
-	env_config();
-
 	/* complain about old configuration settings if they exist */
-	if (exists(RC_CONF_OLD)) {
+	if (access(RC_CONF_OLD, F_OK) == 0) {
 		ewarn("%s still exists on your system and should be removed.", RC_CONF_OLD);
 		ewarn("Please migrate to the appropriate settings in %s", RC_CONF);
 	}
@@ -834,6 +805,8 @@ int main(int argc, char **argv)
 				eerror("runlevel `%s' does not exist", optarg);
 				exit(EXIT_FAILURE);
 			}
+			if (rc_is_user())
+				eerrorx("user services do not currently support --override");
 			if (!set_krunlevel(optarg))
 				exit(EXIT_FAILURE);
 			einfo("Overriding next runlevel to %s", optarg);
@@ -859,17 +832,35 @@ int main(int argc, char **argv)
 		}
 	}
 
+	/* Change dir to / to ensure all init scripts don't use stuff in pwd
+	 * For user services, change to the user's HOME instead. */
+	if (rc_is_user() && !(workingdir = getenv("HOME")))
+		eerrorx("HOME is unset.");
+
+	if (chdir(workingdir) == -1)
+		eerror("chdir: %s", strerror(errno));
+
+	/* Ensure our environment is pure
+	 * Also, add our configuration to it */
+	env_filter();
+	env_config();
+
 	newlevel = argv[optind++];
 	/* To make life easier, we only have the shutdown runlevel as
 	 * nothing really needs to know that we're rebooting.
 	 * But for those that do, you can test against RC_REBOOT. */
 	if (newlevel) {
+		if (rc_is_user() && strcmp(newlevel, RC_LEVEL_SYSINIT) == 0)
+			eerrorx("%s: the sysinit runlevel can't be used for user services.", applet);
+
 		if (strcmp(newlevel, "reboot") == 0) {
 			newlevel = UNCONST(RC_LEVEL_SHUTDOWN);
 			setenv("RC_REBOOT", "YES", 1);
 		}
 	}
 
+	atexit(cleanup);
+
 	/* Enable logging */
 	setenv("EINFO_LOG", "openrc", 1);
 
@@ -912,7 +903,7 @@ int main(int argc, char **argv)
 		strcmp(newlevel, RC_LEVEL_SINGLE) == 0))
 	{
 		going_down = true;
-		if (!exists(RC_KRUNLEVEL))
+		if (access(RC_KRUNLEVEL, F_OK) != 0)
 			set_krunlevel(runlevel);
 		rc_runlevel_set(newlevel);
 		setenv("RC_RUNLEVEL", newlevel, 1);
@@ -975,21 +966,14 @@ int main(int argc, char **argv)
 	if ((main_deptree = _rc_deptree_load(0, &regen)) == NULL)
 		eerrorx("failed to load deptree");
 
-	xasprintf(&deptree_skewed, "%s/clock-skewed", svcdir);
-	if (exists(deptree_skewed))
+	if (faccessat(rc_dirfd(RC_DIR_SVCDIR), "clock-skewed", F_OK, 0) == 0)
 		ewarn("WARNING: clock skew detected!");
-	free(deptree_skewed);
 
 	/* Clean the failed services state dir */
 	clean_failed();
 
-	xasprintf(&rc_stopping, "%s/rc.stopping", svcdir);
-	if (mkdir(rc_stopping, 0755) != 0) {
-		if (errno == EACCES)
-			eerrorx("%s: superuser access required", applet);
-		eerrorx("%s: failed to create stopping dir '%s': %s",
-		    applet, rc_stopping, strerror(errno));
-	}
+	if (mkdirat(rc_dirfd(RC_DIR_SVCDIR), "rc.stopping", 0755) != 0)
+		eerrorx("%s: failed to create stopping dir '%s/rc.stopping': %s", applet, rc_svcdir(), strerror(errno));
 
 	/* Create a list of all services which we could stop (assuming
 	* they won't be active in the new or current runlevel) including
@@ -1070,8 +1054,7 @@ int main(int argc, char **argv)
 	    going_down ? newlevel : runlevel);
 	hook_out = 0;
 
-	rmdir(rc_stopping);
-	free(rc_stopping);
+	unlinkat(rc_dirfd(RC_DIR_SVCDIR), "rc.stopping", AT_REMOVEDIR);
 
 	/* Store the new runlevel */
 	if (newlevel) {
@@ -1088,9 +1071,7 @@ int main(int argc, char **argv)
 		rc_logger_close();
 #endif
 
-	xasprintf(&rc_starting, "%s/rc.starting", svcdir);
-	mkdir(rc_starting, 0755);
-	free(rc_starting);
+	mkdirat(rc_dirfd(RC_DIR_SVCDIR), "rc.starting", 0755);
 
 	rc_plugin_run(RC_HOOK_RUNLEVEL_START_IN, runlevel);
 	hook_out = RC_HOOK_RUNLEVEL_START_OUT;
@@ -1160,12 +1141,8 @@ int main(int argc, char **argv)
 	 * we need to delete them so that they are regenerated again in the
 	 * default runlevel as they may depend on things that are now
 	 * available */
-	if (regen && strcmp(runlevel, bootlevel) == 0) {
-		char *deptree_cache;
-		xasprintf(&deptree_cache, "%s/deptree", svcdir);
-		unlink(deptree_cache);
-		free(deptree_cache);
-	}
+	if (regen && strcmp(runlevel, bootlevel) == 0)
+		unlinkat(rc_dirfd(RC_DIR_SVCDIR), "deptree", 0);
 
 	return EXIT_SUCCESS;
 }
