@@ -26,8 +26,11 @@
 #include <stdlib.h>
 #include <string.h>
 #include <sys/stat.h>
+#include <dirent.h>
+#include <unistd.h>
+#include <fcntl.h>
 
-#define ERRX fprintf (stderr, "out of memory\n"); exit (1)
+#define ERRX do { fprintf (stderr, "out of memory\n"); exit (1); } while (0)
 
 #define UNCONST(a)		((void *)(uintptr_t)(const void *)(a))
 
@@ -91,6 +94,21 @@ RC_UNUSED static char *xstrdup(const char *str)
 	/* NOTREACHED */
 }
 
+RC_UNUSED static FILE *xopen_memstream(char **ptr, size_t *sizeloc)
+{
+	FILE *f = open_memstream(ptr, sizeloc);
+	if (f)
+		return f;
+	ERRX;
+}
+
+RC_UNUSED static void xclose_memstream(FILE *f)
+{
+	fflush(f);
+	if (ferror(f) || fclose(f) != 0)
+		ERRX;
+}
+
 #undef ERRX
 
 /*
@@ -106,18 +124,59 @@ RC_UNUSED static const char *basename_c(const char *path)
 	return (path);
 }
 
-RC_UNUSED static bool exists(const char *pathname)
+RC_UNUSED static FILE *do_fopenat(int dirfd, const char *pathname, int mode)
 {
-	struct stat buf;
+	int fd = openat(dirfd, pathname, mode, 0666);
+	const char *fmode;
+	FILE *fp;
 
-	return (stat(pathname, &buf) == 0);
+	if (fd == -1)
+		return NULL;
+
+	/* O_CREAT and O_TRUNC in modes 'a', 'w+', 'a+' don't
+	 * really matter after the file has been opened.
+	 *
+	 * Some implementations have O_RDONLY defined to zero,
+	 * thus making it impossible to detect, so, we default
+	 * to "r" if no other mask fully matches. */
+	if ((mode & O_RDWR) == O_RDWR)
+		fmode = "r+";
+	else if ((mode & O_WRONLY) == O_WRONLY)
+		fmode = "w";
+	else
+		fmode = "r";
+
+	if (!(fp = fdopen(fd, fmode)))
+		close(fd);
+	return fp;
 }
 
-RC_UNUSED static bool existss(const char *pathname)
+RC_UNUSED static DIR *do_opendirat(int dirfd, const char *pathname)
 {
-	struct stat buf;
+	int fd = openat(dirfd, pathname, O_RDONLY | O_DIRECTORY);
+	DIR *dp;
 
-	return (stat(pathname, &buf) == 0 && buf.st_size != 0);
+	if (fd == -1)
+		return NULL;
+
+	if (!(dp = fdopendir(fd)))
+		close(fd);
+
+	return dp;
+}
+
+RC_UNUSED static DIR *do_dopendir(int dirfd)
+{
+	int fd = dup(dirfd);
+	DIR *dp;
+
+	if (fd == -1)
+		return NULL;
+
+	if (!(dp = fdopendir(fd)))
+		close(fd);
+
+	return dp;
 }
 
 /*
