@@ -2,7 +2,7 @@
   start-stop-daemon
  * Starts, stops, tests and signals daemons
  *
- * This is essentially a ground up re-write of Debians
+ * This is essentially a ground up re-write of Debian's
  * start-stop-daemon for cleaner code and to integrate into our RC
  * system so we can monitor daemons a little.
  */
@@ -21,7 +21,7 @@
 
 #define ONE_MS           1000000
 
-#ifdef __linux__
+#if defined(__linux__) && !defined(_GNU_SOURCE)
 /* For extra SCHED_* defines. */
 # define _GNU_SOURCE
 #endif
@@ -30,6 +30,7 @@
 #include <fcntl.h>
 #include <getopt.h>
 #include <limits.h>
+#include <inttypes.h>
 #include <grp.h>
 #include <pthread.h>
 #include <pwd.h>
@@ -46,6 +47,7 @@
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <sys/types.h>
+#include <inttypes.h>
 #include <sys/wait.h>
 #ifdef __linux__
 # include <sys/syscall.h> /* For io priority */
@@ -70,7 +72,7 @@ static struct pam_conv conv = { NULL, NULL};
 #include "queue.h"
 #include "rc.h"
 #include "misc.h"
-#include "pipes.h"
+#include "rc_exec.h"
 #include "schedules.h"
 #include "_usage.h"
 #include "helpers.h"
@@ -89,11 +91,12 @@ enum {
   LONGOPT_SCHEDULER,
   LONGOPT_SCHEDULER_PRIO,
   LONGOPT_SECBITS,
+  LONGOPT_NOTIFY,
 };
 
 const char *applet = NULL;
 const char *extraopts = NULL;
-const char getoptstring[] = "I:KN:PR:Sa:bc:d:e:g:ik:mn:op:s:tu:r:w:x:1:2:3:4:" \
+const char getoptstring[] = "I:KN:PR:Sa:bc:d:e:g:ik:mn:op:s:tu:r:w:x:0:1:2:3:4:" \
 	getoptstring_COMMON;
 const struct option longopts[] = {
 	{ "capabilities", 1, NULL, LONGOPT_CAPABILITIES},
@@ -123,6 +126,7 @@ const struct option longopts[] = {
 	{ "chroot",       1, NULL, 'r'},
 	{ "wait",         1, NULL, 'w'},
 	{ "exec",         1, NULL, 'x'},
+	{ "stdin",        1, NULL, '0'},
 	{ "stdout",       1, NULL, '1'},
 	{ "stderr",       1, NULL, '2'},
 	{ "stdout-logger",1, NULL, '3'},
@@ -130,6 +134,7 @@ const struct option longopts[] = {
 	{ "progress",     0, NULL, 'P'},
 	{ "scheduler",    1, NULL, LONGOPT_SCHEDULER},
 	{ "scheduler-priority",    1, NULL, LONGOPT_SCHEDULER_PRIO},
+	{ "notify",        1, NULL, LONGOPT_NOTIFY},
 	longopts_COMMON
 };
 const char * const longopts_help[] = {
@@ -160,6 +165,7 @@ const char * const longopts_help[] = {
 	"Chroot to this directory",
 	"Milliseconds to wait for daemon start",
 	"Binary to start/stop",
+	"Redirect stdin to file",
 	"Redirect stdout to file",
 	"Redirect stderr to file",
 	"Redirect stdout to process",
@@ -167,6 +173,7 @@ const char * const longopts_help[] = {
 	"Print dots each second while waiting",
 	"Set process scheduler",
 	"Set process scheduler priority",
+	"Configures experimental notification behaviour",
 	longopts_help_COMMON
 };
 const char *usagestring = NULL;
@@ -316,6 +323,7 @@ int main(int argc, char **argv)
 	gid_t gid = 0;
 	char *home = NULL;
 	int tid = 0;
+	char *redirect_stdin = NULL;
 	char *redirect_stderr = NULL;
 	char *redirect_stdout = NULL;
 	char *stderr_process = NULL;
@@ -353,6 +361,8 @@ int main(int argc, char **argv)
 	int pipefd[2];
 	char readbuf[1];
 	ssize_t ss;
+	struct notify notify = {0};
+	int ret = EXIT_SUCCESS;
 
 	applet = basename_c(argv[0]);
 	atexit(cleanup);
@@ -361,6 +371,9 @@ int main(int argc, char **argv)
 	signal_setup(SIGQUIT, handle_signal);
 	signal_setup(SIGTERM, handle_signal);
 
+	if (rc_yesno(getenv("RC_USER_SERVICES")))
+		rc_set_user();
+
 	openlog(applet, LOG_PID, LOG_DAEMON);
 
 	if ((tmp = getenv("SSD_NICELEVEL")))
@@ -594,6 +607,10 @@ int main(int argc, char **argv)
 			exec = optarg;
 			break;
 
+		case '0':   /* --stdin /path/to/stdin.input-file */
+			redirect_stdin = optarg;
+			break;
+
 		case '1':   /* --stdout /path/to/stdout.lgfile */
 			redirect_stdout = optarg;
 			break;
@@ -618,6 +635,10 @@ int main(int argc, char **argv)
 			sscanf(optarg, "%d", &sched_prio);
 			break;
 
+		case LONGOPT_NOTIFY:
+			notify = notify_parse(svcname ? svcname : applet, optarg);
+			break;
+
 		case_RC_COMMON_GETOPT
 		}
 
@@ -720,7 +741,7 @@ int main(int argc, char **argv)
 					xasprintf(&exec_file, "%s/%s/%s", ch_root, token, exec);
 				else
 					xasprintf(&exec_file, "%s/%s", token, exec);
-				if (exec_file && exists(exec_file))
+				if (exec_file && access(exec_file, F_OK) == 0)
 					break;
 				free(exec_file);
 				exec_file = NULL;
@@ -728,7 +749,7 @@ int main(int argc, char **argv)
 			free(tmp);
 		}
 	}
-	if (start && !exists(exec_file)) {
+	if (start && access(exec_file, F_OK) != 0) {
 		eerror("%s: %s does not exist", applet,
 		    exec_file ? exec_file : exec);
 		free(exec_file);
@@ -804,7 +825,7 @@ int main(int argc, char **argv)
 		 * remove information about it as it may have unexpectedly
 		 * crashed out. We should also return success as the end
 		 * result would be the same. */
-		if (pidfile && exists(pidfile))
+		if (pidfile)
 			unlink(pidfile);
 		if (svcname)
 			rc_service_daemon_set(svcname, exec,
@@ -836,9 +857,9 @@ int main(int argc, char **argv)
 		printf("\n");
 		eindent();
 		if (uid != 0)
-			einfo("as user id %d", uid);
+			einfo("as user id %"PRIuMAX, (uintmax_t)uid);
 		if (gid != 0)
-			einfo("as group id %d", gid);
+			einfo("as group id %"PRIuMAX, (uintmax_t)gid);
 		if (ch_root)
 			einfo("in root `%s'", ch_root);
 		if (ch_dir)
@@ -870,6 +891,8 @@ int main(int argc, char **argv)
 
 	/* Child process - lets go! */
 	if (pid == 0) {
+		gid_t group_buf[32], *group_list = group_buf;
+		int group_count = ARRAY_SIZE(group_buf);
 		pid_t mypid = getpid();
 		close(pipefd[0]); /* Close the read end of the pipe. */
 		umask(numask);
@@ -913,6 +936,29 @@ int main(int argc, char **argv)
 			fclose(fp);
 		}
 
+#ifdef HAVE_PAM
+		if (changeuser != NULL) {
+			pamr = pam_start("start-stop-daemon",
+			    changeuser, &conv, &pamh);
+
+			if (pamr == PAM_SUCCESS)
+				pamr = pam_acct_mgmt(pamh, PAM_SILENT);
+			if (pamr == PAM_SUCCESS)
+				pamr = pam_open_session(pamh, PAM_SILENT);
+			if (pamr != PAM_SUCCESS)
+				eerrorx("%s: pam error: %s",
+					applet, pam_strerror(pamh, pamr));
+		}
+#endif
+		if (changeuser && getgrouplist(changeuser, gid, group_list, &group_count) < 0) {
+			group_list = xmalloc(group_count * sizeof(*group_list));
+			if (getgrouplist(changeuser, gid, group_list, &group_count) < 0)
+				eerrorx("%s: getgrouplist(%s, %"PRIuMAX")", applet, changeuser, (uintmax_t)gid);
+		}
+
+		/* Close any fd's to the passwd database */
+		endpwent();
+
 		if (ch_root && chroot(ch_root) < 0)
 			eerrorx("%s: chroot `%s': %s",
 			    applet, ch_root, strerror(errno));
@@ -930,37 +976,21 @@ int main(int argc, char **argv)
 			fclose(fp);
 		}
 
-#ifdef HAVE_PAM
-		if (changeuser != NULL) {
-			pamr = pam_start("start-stop-daemon",
-			    changeuser, &conv, &pamh);
-
-			if (pamr == PAM_SUCCESS)
-				pamr = pam_acct_mgmt(pamh, PAM_SILENT);
-			if (pamr == PAM_SUCCESS)
-				pamr = pam_open_session(pamh, PAM_SILENT);
-			if (pamr != PAM_SUCCESS)
-				eerrorx("%s: pam error: %s",
-					applet, pam_strerror(pamh, pamr));
-		}
-#endif
 
 		if (gid && setgid(gid))
-			eerrorx("%s: unable to set groupid to %d",
-			    applet, gid);
-		if (changeuser && initgroups(changeuser, gid))
-			eerrorx("%s: initgroups (%s, %d)",
-			    applet, changeuser, gid);
+			eerrorx("%s: unable to set groupid to %"PRIuMAX,
+			    applet, (uintmax_t)gid);
+		if (changeuser && setgroups(group_count, group_list))
+			eerrorx("%s: setgroups() failed", applet);
+		if (group_list != group_buf)
+			free(group_list);
 #ifdef __linux__
 		if (uid && cap_setuid(uid))
 #else
 		if (uid && setuid(uid))
 #endif
-			eerrorx ("%s: unable to set userid to %d",
-			    applet, uid);
-
-		/* Close any fd's to the passwd database */
-		endpwent();
+			eerrorx ("%s: unable to set userid to %"PRIuMAX,
+			    applet, (uintmax_t)uid);
 
 #ifdef __linux__
 		if (cap_iab != NULL) {
@@ -1014,18 +1044,13 @@ int main(int argc, char **argv)
 		}
 #endif
 
+		unsetenv("SVCNAME");
 		TAILQ_FOREACH(env, env_list, entries) {
-			if ((strncmp(env->value, "RC_", 3) == 0 &&
-				strncmp(env->value, "RC_SERVICE=", 11) != 0 &&
-				strncmp(env->value, "RC_SVCNAME=", 11) != 0) ||
-				strncmp(env->value, "SSD_NICELEVEL=", 14) == 0 ||
-				strncmp(env->value, "SSD_IONICELEVEL=", 16) == 0 ||
-				strncmp(env->value, "SSD_OOM_SCORE_ADJ=", 18) == 0)
-			{
-				p = strchr(env->value, '=');
-				*p = '\0';
+			if (strncmp(env->value, "RC_", 3) == 0
+					|| strncmp(env->value, "SSD_", 4) == 0
+					|| strncmp(env->value, "EINFO_", 6) == 0) {
+				*strchr(env->value, '=') = '\0';
 				unsetenv(env->value);
-				continue;
 			}
 		}
 		rc_stringlist_free(env_list);
@@ -1060,6 +1085,13 @@ int main(int argc, char **argv)
 		stdin_fd = devnull_fd;
 		stdout_fd = devnull_fd;
 		stderr_fd = devnull_fd;
+		if (redirect_stdin) {
+			if ((stdin_fd = open(redirect_stdin,
+				    O_RDONLY)) == -1)
+				eerrorx("%s: unable to open the input file"
+				    " for stdin `%s': %s",
+				    applet, redirect_stdin, strerror(errno));
+		}
 		if (redirect_stdout) {
 			if ((stdout_fd = open(redirect_stdout,
 				    O_WRONLY | O_CREAT | O_APPEND,
@@ -1068,7 +1100,7 @@ int main(int argc, char **argv)
 				    " for stdout `%s': %s",
 				    applet, redirect_stdout, strerror(errno));
 		}else if (stdout_process) {
-			stdout_fd = rc_pipe_command(stdout_process);
+			stdout_fd = rc_pipe_command(stdout_process, devnull_fd);
 			if (stdout_fd == -1)
 				eerrorx("%s: unable to open the logging process"
 				    " for stdout `%s': %s",
@@ -1082,7 +1114,7 @@ int main(int argc, char **argv)
 				    " for stderr `%s': %s",
 				    applet, redirect_stderr, strerror(errno));
 		}else if (stderr_process) {
-			stderr_fd = rc_pipe_command(stderr_process);
+			stderr_fd = rc_pipe_command(stderr_process, devnull_fd);
 			if (stderr_fd == -1)
 				eerrorx("%s: unable to open the logging process"
 				    " for stderr `%s': %s",
@@ -1100,6 +1132,13 @@ int main(int argc, char **argv)
 
 		cloexec_fds_from(3);
 
+		if (notify.type == NOTIFY_FD) {
+			if (close(notify.pipe[0]) == -1)
+				eerrorx("%s: failed to close notify pipe[0]: %s", applet, strerror(errno));
+			if (dup2(notify.pipe[1], notify.fd) == -1)
+				eerrorx("%s: failed to initialize notify fd: %s", applet, strerror(errno));
+		}
+
 		if (scheduler != NULL) {
 			int scheduler_index;
 			struct sched_param sched =  {.sched_priority = sched_prio};
@@ -1184,7 +1223,10 @@ int main(int argc, char **argv)
 			start_wait = 0;
 	}
 
-	if (start_wait > 0) {
+	if (notify.type != NOTIFY_NONE) {
+		if (!notify_wait(applet, notify))
+			ret = EXIT_FAILURE;
+	} else if (start_wait > 0) {
 		struct timespec ts;
 		bool alive = false;
 
@@ -1224,6 +1266,6 @@ int main(int argc, char **argv)
 		rc_service_daemon_set(svcname, exec,
 		    (const char *const *)margv, pidfile, true);
 
-	exit(EXIT_SUCCESS);
+	exit(ret);
 	/* NOTREACHED */
 }
